"""
Tests of the neo.core.irregularlysampledsignal.IrregularySampledSignal class
"""

import unittest

import os
import pickle
import warnings
from copy import deepcopy

import numpy as np
import quantities as pq
from numpy.testing import assert_array_equal

from neo.core.dataobject import ArrayDict

try:
    from IPython.lib.pretty import pretty
except ImportError as err:
    HAVE_IPYTHON = False
else:
    HAVE_IPYTHON = True

try:
    import scipy
except ImportError:
    HAVE_SCIPY = False
else:
    HAVE_SCIPY = True

from neo.core.irregularlysampledsignal import IrregularlySampledSignal
from neo.core import Segment
from neo.core.baseneo import MergeError
from neo.test.tools import (
    assert_arrays_almost_equal,
    assert_arrays_equal,
    assert_neo_object_is_compliant,
    assert_same_sub_schema,
    assert_same_attributes,
    assert_same_annotations,
    assert_same_array_annotations,
)


class TestIrregularlySampledSignalConstruction(unittest.TestCase):
    def test_IrregularlySampledSignal_creation_times_units_signal_units(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"anno1": [23], "anno2": ["A"]}
        sig = IrregularlySampledSignal(
            [1.1, 1.5, 1.7] * pq.ms,
            signal=[20.0, 40.0, 60.0] * pq.mV,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        sig.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(sig)

        assert_array_equal(sig.times, [1.1, 1.5, 1.7] * pq.ms)
        assert_array_equal(np.asarray(sig).flatten(), np.array([20.0, 40.0, 60.0]))
        self.assertEqual(sig.units, pq.mV)
        self.assertEqual(sig.name, "test")
        self.assertEqual(sig.description, "tester")
        self.assertEqual(sig.file_origin, "test.file")
        self.assertEqual(sig.annotations["test0"], [1, 2])
        self.assertEqual(sig.annotations["test1"], 1.1)
        self.assertEqual(sig.annotations["test2"], "y1")
        self.assertTrue(sig.annotations["test3"])

        assert_arrays_equal(sig.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(sig.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(sig.array_annotations, ArrayDict)

    def test_IrregularlySampledSignal_creation_units_arg(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"anno1": [23], "anno2": ["A"]}
        sig = IrregularlySampledSignal(
            [1.1, 1.5, 1.7],
            signal=[20.0, 40.0, 60.0],
            units=pq.V,
            time_units=pq.s,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        sig.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(sig)

        assert_array_equal(sig.times, [1.1, 1.5, 1.7] * pq.s)
        assert_array_equal(np.asarray(sig).flatten(), np.array([20.0, 40.0, 60.0]))
        self.assertEqual(sig.units, pq.V)
        self.assertEqual(sig.name, "test")
        self.assertEqual(sig.description, "tester")
        self.assertEqual(sig.file_origin, "test.file")
        self.assertEqual(sig.annotations["test0"], [1, 2])
        self.assertEqual(sig.annotations["test1"], 1.1)
        self.assertEqual(sig.annotations["test2"], "y1")
        self.assertTrue(sig.annotations["test3"])

        assert_arrays_equal(sig.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(sig.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(sig.array_annotations, ArrayDict)

    def test_IrregularlySampledSignal_creation_units_rescale(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"anno1": [23], "anno2": ["A"]}
        sig = IrregularlySampledSignal(
            [1.1, 1.5, 1.7] * pq.s,
            signal=[2.0, 4.0, 6.0] * pq.V,
            units=pq.mV,
            time_units=pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        sig.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(sig)

        assert_array_equal(sig.times, [1100, 1500, 1700] * pq.ms)
        assert_array_equal(np.asarray(sig).flatten(), np.array([2000.0, 4000.0, 6000.0]))
        self.assertEqual(sig.units, pq.mV)
        self.assertEqual(sig.name, "test")
        self.assertEqual(sig.description, "tester")
        self.assertEqual(sig.file_origin, "test.file")
        self.assertEqual(sig.annotations["test0"], [1, 2])
        self.assertEqual(sig.annotations["test1"], 1.1)
        self.assertEqual(sig.annotations["test2"], "y1")
        self.assertTrue(sig.annotations["test3"])

        assert_arrays_equal(sig.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(sig.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(sig.array_annotations, ArrayDict)

    def test_IrregularlySampledSignal_different_lens_ValueError(self):
        times = [1.1, 1.5, 1.7] * pq.ms
        signal = [20.0, 40.0, 60.0, 70.0] * pq.mV
        self.assertRaises(ValueError, IrregularlySampledSignal, times, signal)

    def test_IrregularlySampledSignal_no_signal_units_ValueError(self):
        times = [1.1, 1.5, 1.7] * pq.ms
        signal = [20.0, 40.0, 60.0]
        self.assertRaises(ValueError, IrregularlySampledSignal, times, signal)

    def test_IrregularlySampledSignal_no_time_units_ValueError(self):
        times = [1.1, 1.5, 1.7]
        signal = [20.0, 40.0, 60.0] * pq.mV
        self.assertRaises(ValueError, IrregularlySampledSignal, times, signal)


class TestIrregularlySampledSignalProperties(unittest.TestCase):
    def setUp(self):
        self.times = [np.arange(10.0) * pq.s, np.arange(-100.0, 100.0, 10.0) * pq.ms, np.arange(100) * pq.ns]
        self.data = [
            np.arange(10.0) * pq.nA,
            np.arange(-100.0, 100.0, 10.0) * pq.mV,
            np.random.uniform(size=100) * pq.uV,
        ]
        self.signals = [IrregularlySampledSignal(t, signal=D, testattr="test") for D, t in zip(self.data, self.times)]

    def test__compliant(self):
        for signal in self.signals:
            assert_neo_object_is_compliant(signal)

    def test__t_start_getter(self):
        for signal, times in zip(self.signals, self.times):
            self.assertAlmostEqual(signal.t_start, times[0], delta=1e-15)

    def test__t_stop_getter(self):
        for signal, times in zip(self.signals, self.times):
            self.assertAlmostEqual(signal.t_stop, times[-1], delta=1e-15)

    def test__duration_getter(self):
        for signal, times in zip(self.signals, self.times):
            self.assertAlmostEqual(signal.duration, times[-1] - times[0], delta=1e-15)

    def test__sampling_intervals_getter(self):
        for signal, times in zip(self.signals, self.times):
            assert_arrays_almost_equal(signal.sampling_intervals, np.diff(times), threshold=1e-15)

    def test_IrregularlySampledSignal_repr(self):
        sig = IrregularlySampledSignal(
            [1.1, 1.5, 1.7] * pq.s,
            signal=[2.0, 4.0, 6.0] * pq.V,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
        )
        assert_neo_object_is_compliant(sig)

        if np.__version__.split(".")[:2] > ["1", "13"]:
            # see https://github.com/numpy/numpy/blob/master/doc/release/1.14.0-notes.rst#many
            # -changes-to-array-printing-disableable-with-the-new-legacy-printing-mode
            targ = (
                "<IrregularlySampledSignal(array([[2.],\n       [4.],\n       [6.]]) * V "
                "" + "at times [1.1 1.5 1.7] s)>"
            )
        else:
            targ = (
                "<IrregularlySampledSignal(array([[ 2.],\n       [ 4.],\n       [ 6.]]) "
                "* V " + "at times [ 1.1  1.5  1.7] s)>"
            )
        res = repr(sig)
        self.assertEqual(targ, res)


class TestIrregularlySampledSignalArrayMethods(unittest.TestCase):
    def setUp(self):
        self.data1 = np.arange(10.0)
        self.data1quant = self.data1 * pq.mV
        self.time1 = np.logspace(1, 5, 10)
        self.time1quant = self.time1 * pq.ms
        self.arr_ann = {"anno1": [23], "anno2": ["A"]}
        self.signal1 = IrregularlySampledSignal(
            self.time1quant,
            signal=self.data1quant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
            array_annotations=self.arr_ann,
        )
        self.signal1.segment = Segment()

    def test__compliant(self):
        assert_neo_object_is_compliant(self.signal1)
        self.assertEqual(self.signal1.name, "spam")
        self.assertEqual(self.signal1.description, "eggs")
        self.assertEqual(self.signal1.file_origin, "testfile.txt")
        self.assertEqual(self.signal1.annotations, {"arg1": "test"})
        assert_arrays_equal(self.signal1.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(self.signal1.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(self.signal1.array_annotations, ArrayDict)

    def test__slice_should_return_IrregularlySampledSignal(self):
        result = self.signal1[3:8]
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})

        self.assertEqual(result.size, 5)
        self.assertEqual(result.t_start, self.time1quant[3])
        self.assertEqual(result.t_stop, self.time1quant[7])
        assert_array_equal(self.time1quant[3:8], result.times)
        assert_array_equal(self.data1[3:8].reshape(-1, 1), result.magnitude)

        # Test other attributes were copied over (in this case, defaults)
        self.assertEqual(result.file_origin, self.signal1.file_origin)
        self.assertEqual(result.name, self.signal1.name)
        self.assertEqual(result.description, self.signal1.description)
        self.assertEqual(result.annotations, self.signal1.annotations)
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test__getitem_should_return_single_quantity(self):
        self.assertEqual(self.signal1[0], 0 * pq.mV)
        self.assertEqual(self.signal1[9], 9 * pq.mV)
        self.assertRaises(IndexError, self.signal1.__getitem__, 10)

    def test__getitem_out_of_bounds_IndexError(self):
        self.assertRaises(IndexError, self.signal1.__getitem__, 10)

    def test_comparison_operators(self):
        assert_array_equal(
            self.signal1 >= 5 * pq.mV, np.array([[False, False, False, False, False, True, True, True, True, True]]).T
        )
        assert_array_equal(
            self.signal1 == 5 * pq.mV,
            np.array([[False, False, False, False, False, True, False, False, False, False]]).T,
        )
        assert_array_equal(
            self.signal1 == self.signal1, np.array([[True, True, True, True, True, True, True, True, True, True]]).T
        )

    def test__comparison_as_indexing_single_trace(self):
        self.assertEqual(self.signal1[self.signal1 == 5], [5 * pq.mV])

    def test__comparison_as_indexing_multi_trace(self):
        signal = IrregularlySampledSignal(self.time1quant, np.arange(20).reshape((-1, 2)) * pq.V)
        assert_array_equal(signal[signal < 10], np.array([[0, 2, 4, 6, 8], [1, 3, 5, 7, 9]]).T * pq.V)

    def test__indexing_keeps_order_across_channels(self):
        # AnalogSignals with 10 traces each having 5 samples (eg. data[0] = [0,10,20,30,40])
        data = np.array([range(10), range(10, 20), range(20, 30), range(30, 40), range(40, 50)])
        mask = np.full((5, 10), fill_value=False, dtype=bool)
        # selecting one entry per trace
        mask[[0, 1, 0, 3, 0, 2, 4, 3, 1, 4], range(10)] = True

        signal = IrregularlySampledSignal(np.arange(5) * pq.s, np.array(data) * pq.V)
        assert_array_equal(signal[mask], np.array([[0, 11, 2, 33, 4, 25, 46, 37, 18, 49]]) * pq.V)

    def test__indexing_keeps_order_across_time(self):
        # AnalogSignals with 10 traces each having 5 samples (eg. data[0] = [0,10,20,30,40])
        data = np.array([range(10), range(10, 20), range(20, 30), range(30, 40), range(40, 50)])
        mask = np.full((5, 10), fill_value=False, dtype=bool)
        # selecting two entries per trace
        temporal_ids = [0, 1, 0, 3, 1, 2, 4, 2, 1, 4] + [4, 3, 2, 1, 0, 1, 2, 3, 2, 1]
        mask[temporal_ids, list(range(10)) + list(range(10))] = True

        signal = IrregularlySampledSignal(np.arange(5) * pq.s, np.array(data) * pq.V)
        assert_array_equal(
            signal[mask],
            np.array([[0, 11, 2, 13, 4, 15, 26, 27, 18, 19], [40, 31, 22, 33, 14, 25, 46, 37, 28, 49]]) * pq.V,
        )

    def test__comparison_with_inconsistent_units_should_raise_Exception(self):
        self.assertRaises(ValueError, self.signal1.__gt__, 5 * pq.nA)

    def test_simple_statistics(self):
        targmean = self.signal1[:-1] * np.diff(self.time1quant).reshape(-1, 1)
        targmean = targmean.sum() / (self.time1quant[-1] - self.time1quant[0])
        self.assertEqual(self.signal1.max(), 9 * pq.mV)
        self.assertEqual(self.signal1.min(), 0 * pq.mV)
        self.assertEqual(self.signal1.mean(), targmean)

    def test_mean_interpolation_NotImplementedError(self):
        self.assertRaises(NotImplementedError, self.signal1.mean, True)

    def test__rescale_same(self):
        result = self.signal1.copy()
        result = result.rescale(pq.mV)

        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(result.units, 1 * pq.mV)
        assert_array_equal(result.magnitude, self.data1.reshape(-1, 1))
        assert_array_equal(result.times, self.time1quant)
        assert_same_sub_schema(result, self.signal1)

        self.assertIsInstance(result.segment, Segment)
        self.assertIs(result.segment, self.signal1.segment)

    def test__rescale_new(self):
        result = self.signal1.copy()
        result = result.rescale(pq.uV)

        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(result.units, 1 * pq.uV)
        assert_arrays_almost_equal(np.array(result), self.data1.reshape(-1, 1) * 1000.0, 1e-10)
        assert_array_equal(result.times, self.time1quant)

        self.assertIsInstance(result.segment, Segment)
        self.assertIs(result.segment, self.signal1.segment)

    def test__rescale_new_incompatible_ValueError(self):
        self.assertRaises(ValueError, self.signal1.rescale, pq.nA)

    def test_time_slice(self):
        targdataquant = [[1.0], [2.0], [3.0]] * pq.mV
        targtime = np.logspace(1, 5, 10)
        targtimequant = targtime[1:4] * pq.ms
        targ_signal = IrregularlySampledSignal(
            targtimequant,
            signal=targdataquant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )

        t_start = 15
        t_stop = 250
        result = self.signal1.time_slice(t_start, t_stop)

        assert_array_equal(result, targ_signal)
        assert_array_equal(result.times, targtimequant)
        self.assertEqual(result.units, 1 * pq.mV)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test__time_slice_deepcopy_annotations(self):
        params1 = {"test0": "y1", "test1": ["deeptest"], "test2": True}
        self.signal1.annotate(**params1)

        result = self.signal1.time_slice(None, None)

        # Change annotations of original
        params2 = {"test0": "y2", "test2": False}
        self.signal1.annotate(**params2)
        self.signal1.annotations["test1"][0] = "shallowtest"

        self.assertNotEqual(self.signal1.annotations["test0"], result.annotations["test0"])
        self.assertNotEqual(self.signal1.annotations["test1"], result.annotations["test1"])
        self.assertNotEqual(self.signal1.annotations["test2"], result.annotations["test2"])

        # Change annotations of result
        params3 = {"test0": "y3"}
        result.annotate(**params3)
        result.annotations["test1"][0] = "shallowtest2"

        self.assertNotEqual(self.signal1.annotations["test0"], result.annotations["test0"])
        self.assertNotEqual(self.signal1.annotations["test1"], result.annotations["test1"])
        self.assertNotEqual(self.signal1.annotations["test2"], result.annotations["test2"])

    def test__time_slice_deepcopy_array_annotations(self):
        length = self.signal1.shape[-1]
        params1 = {
            "test0": [f"y{i}" for i in range(length)],
            "test1": ["deeptest" for i in range(length)],
            "test2": [(-1) ** i > 0 for i in range(length)],
        }
        self.signal1.array_annotate(**params1)
        result = self.signal1.time_slice(None, None)

        # Change annotations of original
        params2 = {
            "test0": [f"x{i}" for i in range(length)],
            "test2": [(-1) ** (i + 1) > 0 for i in range(length)],
        }
        self.signal1.array_annotate(**params2)
        self.signal1.array_annotations["test1"][0] = "shallowtest"

        self.assertFalse(all(self.signal1.array_annotations["test0"] == result.array_annotations["test0"]))
        self.assertFalse(all(self.signal1.array_annotations["test1"] == result.array_annotations["test1"]))
        self.assertFalse(all(self.signal1.array_annotations["test2"] == result.array_annotations["test2"]))

        # Change annotations of result
        params3 = {"test0": [f"z{i}" for i in range(1, result.shape[-1] + 1)]}
        result.array_annotate(**params3)
        result.array_annotations["test1"][0] = "shallow2"
        self.assertFalse(all(self.signal1.array_annotations["test0"] == result.array_annotations["test0"]))
        self.assertFalse(all(self.signal1.array_annotations["test1"] == result.array_annotations["test1"]))
        self.assertFalse(all(self.signal1.array_annotations["test2"] == result.array_annotations["test2"]))

    def test__time_slice_deepcopy_data(self):
        result = self.signal1.time_slice(None, None)

        # Change values of original array
        self.signal1[2] = 7.3 * self.signal1.units

        self.assertFalse(all(self.signal1 == result))

        # Change values of sliced array
        result[3] = 9.5 * result.units

        self.assertFalse(all(self.signal1 == result))

    def test_time_slice_out_of_boundries(self):
        targdataquant = self.data1quant
        targtimequant = self.time1quant
        targ_signal = IrregularlySampledSignal(
            targtimequant,
            signal=targdataquant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )

        t_start = 0
        t_stop = 2500000
        result = self.signal1.time_slice(t_start, t_stop)

        assert_array_equal(result, targ_signal)
        assert_array_equal(result.times, targtimequant)
        self.assertEqual(result.units, 1 * pq.mV)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_empty(self):
        targdataquant = [] * pq.mV
        targtimequant = [] * pq.ms
        targ_signal = IrregularlySampledSignal(
            targtimequant,
            signal=targdataquant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )

        t_start = 15
        t_stop = 250
        result = targ_signal.time_slice(t_start, t_stop)

        assert_array_equal(result, targ_signal)
        assert_array_equal(result.times, targtimequant)
        self.assertEqual(result.units, 1 * pq.mV)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        self.assertEqual(result.array_annotations, {})
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_stop(self):
        targdataquant = [[1.0], [2.0], [3.0], [4.0], [5.0], [6.0], [7.0], [8.0], [9.0]] * pq.mV
        targtime = np.logspace(1, 5, 10)
        targtimequant = targtime[1:10] * pq.ms
        targ_signal = IrregularlySampledSignal(
            targtimequant,
            signal=targdataquant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )

        t_start = 15
        t_stop = None
        result = self.signal1.time_slice(t_start, t_stop)

        assert_array_equal(result, targ_signal)
        assert_array_equal(result.times, targtimequant)
        self.assertEqual(result.units, 1 * pq.mV)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_start(self):
        targdataquant = [[0.0], [1.0], [2.0], [3.0]] * pq.mV
        targtime = np.logspace(1, 5, 10)
        targtimequant = targtime[0:4] * pq.ms
        targ_signal = IrregularlySampledSignal(
            targtimequant,
            signal=targdataquant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )

        t_start = None
        t_stop = 250
        result = self.signal1.time_slice(t_start, t_stop)

        assert_array_equal(result, targ_signal)
        assert_array_equal(result.times, targtimequant)
        self.assertEqual(result.units, 1 * pq.mV)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_both(self):
        targdataquant = [[0.0], [1.0], [2.0], [3.0], [4.0], [5.0], [6.0], [7.0], [8.0], [9.0]] * pq.mV
        targtime = np.logspace(1, 5, 10)
        targtimequant = targtime[0:10] * pq.ms
        targ_signal = IrregularlySampledSignal(
            targtimequant,
            signal=targdataquant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )

        t_start = None
        t_stop = None
        result = self.signal1.time_slice(t_start, t_stop)

        assert_array_equal(result, targ_signal)
        assert_array_equal(result.times, targtimequant)
        self.assertEqual(result.units, 1 * pq.mV)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_differnt_units(self):
        targdataquant = [[1.0], [2.0], [3.0]] * pq.mV
        targtime = np.logspace(1, 5, 10)
        targtimequant = targtime[1:4] * pq.ms
        targ_signal = IrregularlySampledSignal(
            targtimequant,
            signal=targdataquant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )

        t_start = 15
        t_stop = 250

        t_start = 0.015 * pq.s
        t_stop = 0.250 * pq.s

        result = self.signal1.time_slice(t_start, t_stop)

        assert_array_equal(result, targ_signal)
        assert_array_equal(result.times, targtimequant)
        self.assertEqual(result.units, 1 * pq.mV)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test__time_slice_should_set_parents_to_None(self):
        # When timeslicing, a deep copy is made,
        # thus the reference to parent objects should be destroyed
        result = self.signal1.time_slice(1 * pq.ms, 3 * pq.ms)
        self.assertEqual(result.segment, None)

    def test__deepcopy_should_set_parents_objects_to_None(self):
        # Deepcopy should destroy references to parents
        result = deepcopy(self.signal1)
        self.assertEqual(result.segment, None)

    def test__time_shift_same_attributes(self):
        result = self.signal1.time_shift(1 * pq.ms)
        assert_same_attributes(result, self.signal1, exclude=["times", "t_start", "t_stop"])

    def test__time_shift_same_annotations(self):
        result = self.signal1.time_shift(1 * pq.ms)
        assert_same_annotations(result, self.signal1)

    def test__time_shift_same_array_annotations(self):
        result = self.signal1.time_shift(1 * pq.ms)
        assert_same_array_annotations(result, self.signal1)

    def test__time_shift_should_set_parents_to_None(self):
        # When time-shifting, a deep copy is made,
        # thus the reference to parent objects should be destroyed
        result = self.signal1.time_shift(1 * pq.ms)
        self.assertEqual(result.segment, None)

    def test__time_shift_by_zero(self):
        shifted = self.signal1.time_shift(0 * pq.ms)
        assert_arrays_equal(shifted.times, self.signal1.times)

    def test__time_shift_same_units(self):
        shifted = self.signal1.time_shift(10 * pq.ms)
        assert_arrays_equal(shifted.times, self.signal1.times + 10 * pq.ms)

    def test__time_shift_different_units(self):
        shifted = self.signal1.time_shift(1 * pq.s)
        assert_arrays_equal(shifted.times, self.signal1.times + 1000 * pq.ms)

    def test_as_array(self):
        sig_as_arr = self.signal1.as_array()
        self.assertIsInstance(sig_as_arr, np.ndarray)
        assert_array_equal(self.data1, sig_as_arr.flat)

    def test_as_quantity(self):
        sig_as_q = self.signal1.as_quantity()
        self.assertIsInstance(sig_as_q, pq.Quantity)
        assert_array_equal(self.data1, sig_as_q.magnitude.flat)

    def test__copy_should_preserve_parent_objects(self):
        result = self.signal1.copy()
        self.assertIs(result.segment, self.signal1.segment)

    @unittest.skipUnless(HAVE_SCIPY, "requires Scipy")
    def test_resample(self):
        factors = [1, 2, 10]
        for factor in factors:
            result = self.signal1.resample(self.signal1.shape[0] * factor)
            np.testing.assert_allclose(self.signal1.magnitude, result.magnitude[::factor], rtol=1e-7, atol=0)


class TestIrregularlySampledSignalCombination(unittest.TestCase):
    def setUp(self):
        self.data1 = np.arange(10.0)
        self.data1quant = self.data1 * pq.mV
        self.time1 = np.logspace(1, 5, 10)
        self.time1quant = self.time1 * pq.ms
        self.arr_ann = {"anno1": [23], "anno2": ["A"]}
        self.signal1 = IrregularlySampledSignal(
            self.time1quant,
            signal=self.data1quant,
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
            array_annotations=self.arr_ann,
        )

    def test__compliant(self):
        assert_neo_object_is_compliant(self.signal1)
        self.assertEqual(self.signal1.name, "spam")
        self.assertEqual(self.signal1.description, "eggs")
        self.assertEqual(self.signal1.file_origin, "testfile.txt")
        self.assertEqual(self.signal1.annotations, {"arg1": "test"})
        assert_arrays_equal(self.signal1.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(self.signal1.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(self.signal1.array_annotations, ArrayDict)

    def test__add_const_quantity_should_preserve_data_complement(self):
        result = self.signal1 + 0.065 * pq.V
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        assert_array_equal(result.magnitude, self.data1.reshape(-1, 1) + 65)
        assert_array_equal(result.times, self.time1quant)
        self.assertEqual(self.signal1[9], 9 * pq.mV)
        self.assertEqual(result[9], 74 * pq.mV)

    def test__add_two_consistent_signals_should_preserve_data_complement(self):
        data2 = np.arange(10.0, 20.0)
        data2quant = data2 * pq.mV
        signal2 = IrregularlySampledSignal(self.time1quant, signal=data2quant)
        assert_neo_object_is_compliant(signal2)

        result = self.signal1 + signal2
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        targ = IrregularlySampledSignal(
            self.time1quant,
            signal=np.arange(10.0, 30.0, 2.0),
            units="mV",
            name="spam",
            description="eggs",
            file_origin="testfile.txt",
            arg1="test",
        )
        assert_neo_object_is_compliant(targ)

        assert_array_equal(result, targ)
        assert_array_equal(self.time1quant, targ.times)
        assert_array_equal(result.times, targ.times)
        assert_same_sub_schema(result, targ)

    def test__add_signals_with_inconsistent_times_AssertionError(self):
        signal2 = IrregularlySampledSignal(self.time1quant * 2.0, signal=np.arange(10.0), units="mV")
        assert_neo_object_is_compliant(signal2)

        self.assertRaises(ValueError, self.signal1.__add__, signal2)

    def test__add_signals_with_inconsistent_dimension_ValueError(self):
        signal2 = np.arange(20).reshape(2, 10)

        self.assertRaises(ValueError, self.signal1.__add__, signal2)

    def test__subtract_const_should_preserve_data_complement(self):
        result = self.signal1 - 65 * pq.mV
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(self.signal1[9], 9 * pq.mV)
        self.assertEqual(result[9], -56 * pq.mV)
        assert_array_equal(result.magnitude, (self.data1 - 65).reshape(-1, 1))
        assert_array_equal(result.times, self.time1quant)

    def test__subtract_from_const_should_return_signal(self):
        result = 10 * pq.mV - self.signal1
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(self.signal1[9], 9 * pq.mV)
        self.assertEqual(result[9], 1 * pq.mV)
        assert_array_equal(result.magnitude, (10 - self.data1).reshape(-1, 1))
        assert_array_equal(result.times, self.time1quant)

    def test__mult_signal_by_const_float_should_preserve_data_complement(self):
        result = self.signal1 * 2.0
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(self.signal1[9], 9 * pq.mV)
        self.assertEqual(result[9], 18 * pq.mV)
        assert_array_equal(result.magnitude, self.data1.reshape(-1, 1) * 2)
        assert_array_equal(result.times, self.time1quant)

    def test__mult_signal_by_const_array_should_preserve_data_complement(self):
        result = self.signal1 * np.array(2.0)
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(self.signal1[9], 9 * pq.mV)
        self.assertEqual(result[9], 18 * pq.mV)
        assert_array_equal(result.magnitude, self.data1.reshape(-1, 1) * 2)
        assert_array_equal(result.times, self.time1quant)

    def test__divide_signal_by_const_should_preserve_data_complement(self):
        result = self.signal1 / 0.5
        self.assertIsInstance(result, IrregularlySampledSignal)
        assert_neo_object_is_compliant(result)
        self.assertEqual(result.name, "spam")
        self.assertEqual(result.description, "eggs")
        self.assertEqual(result.file_origin, "testfile.txt")
        self.assertEqual(result.annotations, {"arg1": "test"})
        assert_arrays_equal(result.array_annotations["anno1"], np.array([23]))
        assert_arrays_equal(result.array_annotations["anno2"], np.array(["A"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(self.signal1[9], 9 * pq.mV)
        self.assertEqual(result[9], 18 * pq.mV)
        assert_array_equal(result.magnitude, self.data1.reshape(-1, 1) / 0.5)
        assert_array_equal(result.times, self.time1quant)

    @unittest.skipUnless(HAVE_IPYTHON, "requires IPython")
    def test__pretty(self):
        res = pretty(self.signal1)
        signal = self.signal1
        targ = (
            (
                f"IrregularlySampledSignal with {signal.shape[1]} channels of length {signal.shape[0]}; units {signal.units.dimensionality.unicode}; datatype {signal.dtype}\n"
                ""
            )
            + (f"name: '{signal.name}'\ndescription: '{signal.description}'\n")
            + (f"annotations: {str(signal.annotations)}\n")
            + (f"sample times: {signal.times[:10]}")
        )
        self.assertEqual(res, targ)

    def test__merge(self):
        data1 = np.arange(1000.0, 1066.0).reshape((11, 6)) * pq.uV
        data2 = np.arange(2.0, 2.033, 0.001).reshape((11, 3)) * pq.mV
        times1 = np.arange(11.0) * pq.ms
        times2 = np.arange(1.0, 12.0) * pq.ms
        arr_ann1 = {"anno1": np.arange(6), "anno2": ["a", "b", "c", "d", "e", "f"]}
        arr_ann2 = {"anno1": np.arange(100, 103), "anno3": []}

        signal1 = IrregularlySampledSignal(
            times1,
            data1,
            name="signal1",
            description="test signal",
            file_origin="testfile.txt",
            array_annotations=arr_ann1,
        )
        signal2 = IrregularlySampledSignal(
            times1,
            data2,
            name="signal2",
            description="test signal",
            file_origin="testfile.txt",
            array_annotations=arr_ann2,
        )
        signal3 = IrregularlySampledSignal(
            times2, data2, name="signal3", description="test signal", file_origin="testfile.txt"
        )

        with warnings.catch_warnings(record=True) as w:
            merged12 = signal1.merge(signal2)

            self.assertTrue(len(w) == 1)
            self.assertEqual(w[0].category, UserWarning)
            self.assertSequenceEqual(
                str(w[0].message),
                "The following array annotations were "
                "omitted, because they were only present"
                " in one of the merged objects: "
                "['anno2'] from the one that was merged "
                "into and ['anno3'] from the one that "
                "was merged into the other",
            )

        target_data12 = np.hstack([data1, data2.rescale(pq.uV)])

        assert_neo_object_is_compliant(signal1)
        assert_neo_object_is_compliant(signal2)
        assert_neo_object_is_compliant(merged12)

        self.assertAlmostEqual(merged12[5, 0], 1030.0 * pq.uV, 9)
        self.assertAlmostEqual(merged12[5, 6], 2015.0 * pq.uV, 9)

        self.assertEqual(merged12.name, "merge(signal1, signal2)")
        self.assertEqual(merged12.file_origin, "testfile.txt")
        assert_arrays_equal(merged12.array_annotations["anno1"], np.array([0, 1, 2, 3, 4, 5, 100, 101, 102]))
        self.assertIsInstance(merged12.array_annotations, ArrayDict)

        assert_arrays_equal(merged12.magnitude, target_data12)

        self.assertRaises(MergeError, signal1.merge, signal3)

    def test_concatenate_simple(self):
        signal1 = IrregularlySampledSignal(signal=[0, 1, 2, 3] * pq.s, times=[1, 10, 11, 14] * pq.s)
        signal2 = IrregularlySampledSignal(signal=[4, 5, 6] * pq.s, times=[15, 16, 21] * pq.s)

        result = signal1.concatenate(signal2)
        assert_array_equal(np.array([0, 1, 2, 3, 4, 5, 6]).reshape((-1, 1)), result.magnitude)
        assert_array_equal(np.array([1, 10, 11, 14, 15, 16, 21]), result.times)
        for attr in signal1._necessary_attrs:
            if attr[0] == "times":
                continue
            self.assertEqual(getattr(signal1, attr[0], None), getattr(result, attr[0], None))

    def test_concatenate_no_overlap(self):
        signal1 = IrregularlySampledSignal(signal=[0, 1, 2, 3] * pq.s, times=range(4) * pq.s)
        signal2 = IrregularlySampledSignal(signal=[4, 5, 6] * pq.s, times=range(4, 7) * pq.s)

        for allow_overlap in [True, False]:
            result = signal1.concatenate(signal2, allow_overlap=allow_overlap)
            assert_array_equal(np.arange(7).reshape((-1, 1)), result.magnitude)
            assert_array_equal(np.arange(7), result.times)

    def test_concatenate_overlap_exception(self):
        signal1 = IrregularlySampledSignal(signal=[0, 1, 2, 3] * pq.s, times=range(4) * pq.s)
        signal2 = IrregularlySampledSignal(signal=[4, 5, 6] * pq.s, times=range(2, 5) * pq.s)

        self.assertRaises(ValueError, signal1.concatenate, signal2, allow_overlap=False)

    def test_concatenate_overlap(self):
        signal1 = IrregularlySampledSignal(signal=[0, 1, 2, 3] * pq.s, times=range(4) * pq.s)
        signal2 = IrregularlySampledSignal(signal=[4, 5, 6] * pq.s, times=range(2, 5) * pq.s)

        result = signal1.concatenate(signal2, allow_overlap=True)
        assert_array_equal(np.array([0, 1, 2, 4, 3, 5, 6]).reshape((-1, 1)), result.magnitude)
        assert_array_equal(np.array([0, 1, 2, 2, 3, 3, 4]), result.times)

    def test_concatenate_multi_trace(self):
        data1 = np.arange(4).reshape(2, 2)
        data2 = np.arange(4, 8).reshape(2, 2)
        n1 = len(data1)
        n2 = len(data2)
        signal1 = IrregularlySampledSignal(signal=data1 * pq.s, times=range(n1) * pq.s)
        signal2 = IrregularlySampledSignal(signal=data2 * pq.s, times=range(n1, n1 + n2) * pq.s)

        result = signal1.concatenate(signal2)
        data_expected = np.array([[0, 1], [2, 3], [4, 5], [6, 7]])
        assert_array_equal(data_expected, result.magnitude)

    def test_concatenate_array_annotations(self):
        array_anno1 = {"first": ["a", "b"]}
        array_anno2 = {"first": ["a", "b"], "second": ["c", "d"]}
        data1 = np.arange(4).reshape(2, 2)
        data2 = np.arange(4, 8).reshape(2, 2)
        n1 = len(data1)
        n2 = len(data2)
        signal1 = IrregularlySampledSignal(signal=data1 * pq.s, times=range(n1) * pq.s, array_annotations=array_anno1)
        signal2 = IrregularlySampledSignal(
            signal=data2 * pq.s, times=range(n1, n1 + n2) * pq.s, array_annotations=array_anno2
        )

        result = signal1.concatenate(signal2)
        assert_array_equal(array_anno1.keys(), result.array_annotations.keys())

        for k in array_anno1.keys():
            assert_array_equal(np.asarray(array_anno1[k]), result.array_annotations[k])


class TestAnalogSignalFunctions(unittest.TestCase):
    def test__pickle(self):
        signal1 = IrregularlySampledSignal(np.arange(10.0) / 100 * pq.s, np.arange(10.0), units="mV")

        fobj = open("./pickle", "wb")
        pickle.dump(signal1, fobj)
        fobj.close()

        fobj = open("./pickle", "rb")
        try:
            signal2 = pickle.load(fobj)
        except ValueError:
            signal2 = None

        assert_array_equal(signal1, signal2)
        fobj.close()
        os.remove("./pickle")


class TestIrregularlySampledSignalEquality(unittest.TestCase):
    def test__signals_with_different_times_should_be_not_equal(self):
        signal1 = IrregularlySampledSignal(np.arange(10.0) / 100 * pq.s, np.arange(10.0), units="mV")
        signal2 = IrregularlySampledSignal(np.arange(10.0) / 100 * pq.ms, np.arange(10.0), units="mV")
        self.assertNotEqual(signal1, signal2)


if __name__ == "__main__":
    unittest.main()



"""
This module implements generic container base class that all neo container
object inherit from.  It provides shared methods for all container types.

:class:`Container` is derived from :class:`BaseNeo`
"""

from copy import deepcopy

from neo.core import filters
from neo.core.baseneo import BaseNeo, _reference_name, _container_name
from neo.core.objectlist import ObjectList
from neo.core.spiketrain import SpikeTrain
from neo.core.spiketrainlist import SpikeTrainList


def unique_objs(objs):
    """
    Return a list of objects in the list objs where all objects are unique
    using the "is" test.
    """
    seen = set()
    return [obj for obj in objs if id(obj) not in seen and not seen.add(id(obj))]


def filterdata(data, targdict=None, objects=None, **kwargs):
    """
    Return a list of the objects in data matching *any* of the search terms
    in either their attributes or annotations.  Search terms can be
    provided as keyword arguments or a dictionary, either as a positional
    argument after data or to the argument targdict.
    A key of a provided dictionary is the name of the requested annotation
    and the value is a FilterCondition object.
    E.g.: Equal(x), LessThan(x), InRange(x, y).

    targdict can also
    be a list of dictionaries, in which case the filters are applied
    sequentially.

    A list of dictionaries is handled as follows: [ { or } and { or } ]
    If targdict and kwargs are both supplied, the
    targdict filters are applied first, followed by the kwarg filters.
    A targdict of None or {} corresponds to no filters applied, therefore
    returning all child objects. Default targdict is None.
    """

    # if objects are specified, get the classes
    if objects:
        if hasattr(objects, "lower") or isinstance(objects, type):
            objects = [objects]
    elif objects is not None:
        return []

    # handle cases with targdict
    if targdict is None:
        targdict = kwargs
    elif not kwargs:
        pass
    elif hasattr(targdict, "keys"):
        targdict = [targdict, kwargs]
    else:
        targdict += [kwargs]

    if not targdict:
        results = data

    # if multiple dicts are provided, apply each filter sequentially
    elif not hasattr(targdict, "keys"):
        # for performance reasons, only do the object filtering on the first
        # iteration
        results = filterdata(data, targdict=targdict[0], objects=objects)
        for targ in targdict[1:]:
            results = filterdata(results, targdict=targ)
        return results
    else:
        # do the actual filtering
        results = []
        for obj in data:
            for key, value in sorted(targdict.items()):
                if hasattr(obj, key) and getattr(obj, key) == value:
                    results.append(obj)
                    break
                if isinstance(value, filters.FilterCondition) and key in obj.annotations:
                    if value.evaluate(obj.annotations[key]):
                        results.append(obj)
                        break
                if key in obj.annotations and obj.annotations[key] == value:
                    results.append(obj)
                    break

    # remove duplicates from results
    results = list({id(res): res for res in results}.values())

    # keep only objects of the correct classes
    if objects:
        results = [result for result in results if result.__class__ in objects or result.__class__.__name__ in objects]

    if results and all(isinstance(obj, SpikeTrain) for obj in results):
        return SpikeTrainList(results)
    else:
        return results


class Container(BaseNeo):
    """
    This is the base class from which Neo container objects inherit.  It
    derives from :class:`BaseNeo`.

    In addition to the setup :class:`BaseNeo` does, this class also
    automatically sets up the lists to hold the children of the object.

    Each class can define one or more of the following class attributes
    (in  addition to those of BaseNeo):
        :_container_child_objects: Neo container objects that can be children
                                   of this object. An instance attribute
                                   named class.__name__.lower()+'s' will be
                                   automatically defined to hold this child and
                                   will be initialized to an empty list.
        :_data_child_objects: Neo data objects that can be children
                              of this object. An instance attribute named
                              class.__name__.lower()+'s' will be automatically
                              defined to hold this child and will be
                              initialized to an empty list.
        :_repr_pretty_containers: The names of containers attributes printed
                                  when pretty-printing using IPython.

    The following helper properties are available
    (in  addition to those of BaseNeo):
        :_child_objects: All neo objects that can be children of this object.
                                :_container_child_objects: +
                                :_data_child_objects:
        :_container_child_containers: The names of the container attributes
                                      used to store :_container_child_objects:
        :_data_child_containers: The names of the container attributes used
                                 to store :_data_child_objects:
        :_single_child_containers: The names of the container attributes used
                                   to store :_single_child_objects:
        :_child_containers: All child container attributes. Same as :_single_child_containers:
        :_single_children: All objects that are children of the current object
                           where the child can only have one parent of
                           this type.
        :data_children: All data objects that are children of
                        the current object.
        :container_children: All container objects that are children of
                             the current object.
        :children: All Neo objects that are children of the current object.
        :data_children_recur: All data objects that are children of
                              the current object or any of its children,
                              any of its children's children, etc.
        :container_children_recur: All container objects that are children of
                                   the current object or any of its children,
                                   any of its children's children, etc.
        :children_recur: All Neo objects that are children of
                         the current object or any of its children,
                         any of its children's children, etc.

    The following "universal" methods are available
    (in  addition to those of BaseNeo):
        :size: A dictionary where each key is an attribute storing child
               objects and the value is the number of objects stored in that
               attribute.

        :filter(**args): Retrieves children of the current object that
                         have particular properties.

        :list_children_by_class(**args): Retrieves all children of the current
                                         object recursively that are of a
                                         particular class.

        :create_relationship(**args): For each child of the current
                                      object, set its parent
                                      to be the current object.

        :merge(**args): Annotations are merged based on the rules of
                        :merge_annotations:.  Child objects with the same name
                        and a :merge: method are merged using that method.
                        Other child objects are appended to the relevant
                        container attribute.  Parents attributes are NOT
                        changed in this operation.
                        Unlike :BaseNeo.merge:, this method implements
                        all necessary merge rules for a container class.


    Each child class should:
        0) call Container.__init__(self, name=name, description=description,
                                   file_origin=file_origin, **annotations)
           with the universal recommended arguments, plus optional annotations
        1) process its required arguments in its __new__ or __init__ method
        2) process its non-universal recommended arguments (in its __new__ or
           __init__ method
    """

    # Child objects that are a container
    _container_child_objects = ()
    # Child objects that have data
    _data_child_objects = ()
    # Containers that are listed when pretty-printing
    _repr_pretty_containers = ()

    def __init__(self, name=None, description=None, file_origin=None, **annotations):
        """
        Initialize a new :class:`Container` instance.
        """
        super().__init__(name=name, description=description, file_origin=file_origin, **annotations)

    def _get_object_list(self, name):
        """
        Return the container's ObjectList with the given (private) attribute name

        Example:
        >>> segment._get_object_list("_analogsignals")
        """
        return getattr(self, name)

    def _set_object_list(self, name, value):
        """
        Set the contents of the container's ObjectList with the given (private) attribute name

        Example:
        >>> segment._set_object_list("_analogsignals", [sig1, sig2])
        """
        if isinstance(value, list):
            object_list = getattr(self, name)
            object_list.clear()
            object_list.extend(value)
        elif isinstance(value, ObjectList):  # from __iadd__
            setattr(self, name, value)
        else:
            TypeError("value must be a list or an ObjectList")

    @property
    def _child_objects(self):
        """
        Return the names of the classes that can be children of this container.
        """
        return self._container_child_objects + self._data_child_objects

    @property
    def _container_child_containers(self):
        """
        Containers for child objects that are a container and
        have a single parent.
        """
        return tuple([_container_name(child) for child in self._container_child_objects])

    @property
    def _data_child_containers(self):
        """
        Containers for child objects that have data and have a single parent.
        """
        # the following construction removes the duplicate 'regionsofinterest'
        # while preserving the child order (which `set()` would not do)
        # I don't know if preserving the order is important, but I'm playing it safe
        return tuple({_container_name(child): None for child in self._data_child_objects}.keys())

    @property
    def _child_containers(self):
        """
        Containers for child objects with a single parent.
        """
        return tuple({_container_name(child): None for child in self._child_objects}.keys())

    @property
    def _single_children(self):
        """
        All child objects that can only have single parents.
        """
        childs = [list(getattr(self, attr)) for attr in self._child_containers]
        return tuple(sum(childs, []))

    @property
    def data_children(self):
        """
        All data child objects stored in the current object.
        Not recursive.
        """
        childs = [list(getattr(self, attr)) for attr in self._data_child_containers]
        return tuple(sum(childs, []))

    @property
    def container_children(self):
        """
        All container child objects stored in the current object.
        Not recursive.
        """
        childs = [list(getattr(self, attr)) for attr in self._container_child_containers]
        return tuple(sum(childs, []))

    @property
    def children(self):
        """
        All child objects stored in the current object.
        Not recursive.
        """
        return self.data_children + self.container_children

    @property
    def data_children_recur(self):
        """
        All data child objects stored in the current object,
        obtained recursively.
        """
        childs = [list(child.data_children_recur) for child in self.container_children]
        return self.data_children + tuple(sum(childs, []))

    @property
    def container_children_recur(self):
        """
        All container child objects stored in the current object,
        obtained recursively.
        """
        childs = [list(child.container_children_recur) for child in self.container_children]
        return self.container_children + tuple(sum(childs, []))

    @property
    def children_recur(self):
        """
        All child objects stored in the current object,
        obtained recursively.
        """
        return self.data_children_recur + self.container_children_recur

    @property
    def size(self):
        """
        Get dictionary containing the names of child containers in the current
        object as keys and the number of children of that type as values.
        """
        return {name: len(getattr(self, name)) for name in self._child_containers}

    @property
    def _container_lookup(self):
        return {
            cls_name: getattr(self, container_name)
            for cls_name, container_name in zip(self._child_objects, self._child_containers)
        }

    def _get_container(self, cls):
        if hasattr(cls, "proxy_for"):
            cls = cls.proxy_for
        return self._container_lookup[cls.__name__]

    def add(self, *objects):
        """Add a new Neo object to the Container"""
        for obj in objects:
            if obj.__class__.__name__ in self._child_objects or (
                hasattr(obj, "proxy_for") and obj.proxy_for.__name__ in self._child_objects
            ):
                container = self._get_container(obj.__class__)
                container.append(obj)
            else:
                raise TypeError(
                    f"Cannot add object of type {obj.__class__.__name__} "
                    f"to a {self.__class__.__name__}, can only add objects of the "
                    f"following types: {self._child_objects}"
                )

    def filter(self, targdict=None, data=True, container=False, recursive=True, objects=None, **kwargs):
        """
        Return a list of child objects matching *any* of the search terms
        in either their attributes or annotations.  Search terms can be
        provided as keyword arguments or a dictionary, either as a positional
        argument after data or to the argument targdict.
        A key of a provided dictionary is the name of the requested annotation
        and the value is a FilterCondition object.
        E.g.: equal(x), less_than(x), InRange(x, y).

        targdict can also
        be a list of dictionaries, in which case the filters are applied
        sequentially.

        A list of dictionaries is handled as follows: [ { or } and { or } ]
        If targdict and kwargs are both supplied, the
        targdict filters are applied first, followed by the kwarg filters.
        A targdict of None or {} corresponds to no filters applied, therefore
        returning all child objects. Default targdict is None.

        If data is True (default), include data objects.
        If container is True (default False), include container objects.
        If recursive is True (default), descend into child containers for
        objects.

        objects (optional) should be the name of a Neo object type,
        a neo object class, or a list of one or both of these.  If specified,
        only these objects will be returned. If not specified any type of
        object is  returned. Default is None.
        Note that if recursive is True, containers not in objects will still
        be descended into. This overrides data and container.


        Examples::

            >>> obj.filter(name="Vm")
            >>> obj.filter(objects=neo.SpikeTrain)
            >>> obj.filter(targdict={'myannotation':3})
            >>> obj.filter(name=neo.core.filters.Equal(5))
            >>> obj.filter({'name': neo.core.filters.LessThan(5)})
        """

        if isinstance(targdict, str):
            raise TypeError("filtering is based on key-value pairs." " Only a single string was provided.")

        # if objects are specified, get the classes
        if objects:
            data = True
            container = True

        children = []
        # get the objects we want
        if data:
            if recursive:
                children.extend(self.data_children_recur)
            else:
                children.extend(self.data_children)
        if container:
            if recursive:
                children.extend(self.container_children_recur)
            else:
                children.extend(self.container_children)

        filtered = filterdata(children, objects=objects, targdict=targdict, **kwargs)
        if objects == SpikeTrain:
            return SpikeTrainList(items=filtered)
        else:
            return filtered

    def list_children_by_class(self, cls):
        """
        List all children of a particular class recursively.

        You can either provide a class object, a class name,
        or the name of the container storing the class.
        """
        if not hasattr(cls, "lower"):
            cls = cls.__name__
        container_name = _container_name(cls)
        objs = list(getattr(self, container_name, []))
        for child in self.container_children_recur:
            objs.extend(getattr(child, container_name, []))
        return objs

    def check_relationships(self, recursive=True):
        """
        Check that the expected child-parent relationships exist.
        """
        parent_name = _reference_name(self.__class__.__name__)
        for child in self._single_children:
            if hasattr(child, "proxy_for"):
                container = getattr(self, _container_name(child.proxy_for.__name__))
            else:
                container = getattr(self, _container_name(child.__class__.__name__))
            if container.parent is not None:
                if getattr(child, parent_name, None) is not self:
                    raise AttributeError("Child should have its parent as an attribute")
        if recursive:
            for child in self.container_children:
                child.check_relationships(recursive=True)

    def create_relationship(self, force=False, recursive=True):
        """
        For each child of the current object that can only have a single
        parent, set its parent to be the current object.
        For children of the current object, put the current object in the parent list.

        If force is True overwrite any existing relationships
        If recursive is True descend into child objects and create
        relationships there
        """
        parent_name = _reference_name(self.__class__.__name__)
        for child in self._single_children:
            if hasattr(child, parent_name) and getattr(child, parent_name) is None or force:
                setattr(child, parent_name, self)
        if recursive:
            for child in self.container_children:
                child.create_relationship(force=force, recursive=True)

    def __deepcopy__(self, memo):
        """
        Creates a deep copy of the container.
        All contained objects will also be deep copied and relationships
        between all objects will be identical to the original relationships.
        Attributes and annotations of the container are deep copied as well.

        :param memo: (dict) Objects that have been deep copied already
        :return: (Container) Deep copy of input Container
        """
        cls = self.__class__
        necessary_attrs = {}
        for k in self._necessary_attrs:
            necessary_attrs[k[0]] = getattr(self, k[0], None)
        new_container = cls(**necessary_attrs)
        new_container.__dict__.update(self.__dict__)
        memo[id(self)] = new_container
        for k, v in self.__dict__.items():
            try:
                setattr(new_container, k, deepcopy(v, memo))
            except TypeError:
                setattr(new_container, k, v)

        new_container.create_relationship()

        return new_container

    def merge(self, other):
        """
        Merge the contents of another object into this one.

        Container children of the current object with the same name will be
        merged.  All other objects will be appended to the list of objects
        in this one.  Duplicate copies of the same object will be skipped.

        Annotations are merged such that only items not present in the current
        annotations are added.

        Note that the other object will be linked inconsistently to other Neo objects
        after the merge operation and should not be used further.
        """
        # merge containers with the same name
        for container in self._container_child_containers:
            lookup = {obj.name: obj for obj in getattr(self, container)}
            ids = [id(obj) for obj in getattr(self, container)]
            for obj in getattr(other, container):
                if id(obj) in ids:
                    continue
                if obj.name in lookup:
                    lookup[obj.name].merge(obj)
                else:
                    lookup[obj.name] = obj
                    ids.append(id(obj))
                    getattr(self, container).append(obj)

        # for data objects, ignore the name and just add them
        for container in self._data_child_containers:
            objs = getattr(self, container)
            lookup = {obj.name: i for i, obj in enumerate(objs)}
            ids = [id(obj) for obj in objs]
            for obj in getattr(other, container):
                if id(obj) in ids:
                    pass
                elif hasattr(obj, "merge") and obj.name is not None and obj.name in lookup:
                    ind = lookup[obj.name]
                    try:
                        newobj = getattr(self, container)[ind].merge(obj)
                        getattr(self, container)[ind] = newobj
                    except NotImplementedError:
                        getattr(self, container).append(obj)
                        ids.append(id(obj))
                else:
                    lookup[obj.name] = obj
                    ids.append(id(obj))
                    getattr(self, container).append(obj)
                obj.set_parent(self)

        # use the BaseNeo merge as well
        super().merge(other)

    def _repr_pretty_(self, pp, cycle):
        """
        Handle pretty-printing.
        """
        pp.text(self.__class__.__name__)
        pp.text(" with ")

        vals = []
        for container in self._child_containers:
            objs = getattr(self, container)
            if objs:
                vals.append(f"{objs} {container}")
        pp.text(", ".join(vals))

        if self._has_repr_pretty_attrs_():
            pp.breakable()
            self._repr_pretty_attrs_(pp, cycle)

        for container in self._repr_pretty_containers:
            pp.breakable()
            objs = getattr(self, container)
            pp.text(f"# {container} (N={objs})")
            for i, obj in enumerate(objs):
                pp.breakable()
                pp.text(f"{i}: ")
                with pp.indent(3):
                    pp.pretty(obj)



"""
Class for reading data from a 3-brain Biocam system.

See:
https://www.3brain.com/products/single-well/biocam-x

Authors: Alessio Buccino, Robert Wolff
"""

import json
import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class BiocamRawIO(BaseRawIO):
    """
    Class for reading data from a Biocam h5 file.

    Parameters
    ----------
    filename: str, default: ''
        The *.h5 file to be read

    Examples
    --------
        >>> import neo.rawio
        >>> r = neo.rawio.BiocamRawIO(filename='biocam.h5')
        >>> r.parse_header()
        >>> print(r)
        >>> raw_chunk = r.get_analogsignal_chunk(block_index=0,
                                                 seg_index=0,
                                                 i_start=0,
                                                 i_stop=1024,
                                                 channel_names=channel_names)
        >>> float_chunk = r.rescale_signal_raw_to_float(raw_chunk,
                                                        dtype='float64',
                                                        channel_indexes=[0, 3, 6])
    """

    extensions = ["h5", "brw"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        self._header_dict = open_biocam_file_header(self.filename)
        self._num_channels = self._header_dict["num_channels"]
        self._num_frames = self._header_dict["num_frames"]
        self._sampling_rate = self._header_dict["sampling_rate"]
        self._filehandle = self._header_dict["file_handle"]
        self._read_function = self._header_dict["read_function"]
        self._channels = self._header_dict["channels"]
        gain = self._header_dict["gain"]
        offset = self._header_dict["offset"]

        signal_streams = np.array([("Signals", "0")], dtype=_signal_stream_dtype)

        sig_channels = []
        for c, chan in enumerate(self._channels):
            ch_name = f"ch{chan[0]}-{chan[1]}"
            chan_id = str(c + 1)
            sr = self._sampling_rate  # Hz
            dtype = "uint16"
            units = "uV"
            gain = gain
            offset = offset
            stream_id = "0"
            sig_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))
        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        all_starts = [[0.0]]
        return all_starts[block_index][seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._num_frames / self._sampling_rate
        all_stops = [[t_stop]]
        return all_stops[block_index][seg_index]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._num_frames

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index must be 0")
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._num_frames
        if channel_indexes is None:
            channel_indexes = slice(None)
        data = self._read_function(self._filehandle, i_start, i_stop, self._num_channels)
        return data[:, channel_indexes]


def open_biocam_file_header(filename):
    """Open a Biocam hdf5 file, read and return the recording info, pick the correct method to access raw data,
    and return this to the caller."""
    import h5py

    rf = h5py.File(filename, "r")

    if "3BRecInfo" in rf.keys():  # brw v3.x
        # Read recording variables
        rec_vars = rf.require_group("3BRecInfo/3BRecVars/")
        bit_depth = rec_vars["BitDepth"][0]
        max_uv = rec_vars["MaxVolt"][0]
        min_uv = rec_vars["MinVolt"][0]
        num_frames = rec_vars["NRecFrames"][0]
        sampling_rate = rec_vars["SamplingRate"][0]
        signal_inv = rec_vars["SignalInversion"][0]

        # Get the actual number of channels used in the recording
        file_format = rf["3BData"].attrs.get("Version", None)
        format_100 = False
        if file_format == 100:
            num_channels = len(rf["3BData/Raw"][0])
            format_100 = True
        elif file_format in (101, 102) or file_format is None:
            num_channels = int(rf["3BData/Raw"].shape[0] / num_frames)
        else:
            raise Exception("Unknown data file format.")

        # # get channels
        channels = rf["3BRecInfo/3BMeaStreams/Raw/Chs"][:]

        # determine correct function to read data
        if format_100:
            if signal_inv == 1:
                read_function = readHDF5t_100
            elif signal_inv == -1:
                read_function = readHDF5t_100_i
            else:
                raise Exception("Unknown signal inversion")
        else:
            if signal_inv == 1:
                read_function = readHDF5t_101
            elif signal_inv == -1:
                read_function = readHDF5t_101_i
            else:
                raise Exception("Unknown signal inversion")

        gain = (max_uv - min_uv) / (2**bit_depth)
        offset = min_uv

        return dict(
            file_handle=rf,
            num_frames=num_frames,
            sampling_rate=sampling_rate,
            num_channels=num_channels,
            channels=channels,
            file_format=file_format,
            signal_inv=signal_inv,
            read_function=read_function,
            gain=gain,
            offset=offset,
        )
    else:  # brw v4.x
        # Read recording variables
        experiment_settings = json.JSONDecoder().decode(rf["ExperimentSettings"][0].decode())
        max_uv = experiment_settings["ValueConverter"]["MaxAnalogValue"]
        min_uv = experiment_settings["ValueConverter"]["MinAnalogValue"]
        max_digital = experiment_settings["ValueConverter"]["MaxDigitalValue"]
        min_digital = experiment_settings["ValueConverter"]["MinDigitalValue"]
        scale_factor = experiment_settings["ValueConverter"]["ScaleFactor"]
        sampling_rate = experiment_settings["TimeConverter"]["FrameRate"]

        for key in rf:
            if key[:5] == "Well_":
                num_channels = len(rf[key]["StoredChIdxs"])
                if len(rf[key]["Raw"]) % num_channels:
                    raise RuntimeError(f"Length of raw data array is not multiple of channel number in {key}")
                num_frames = len(rf[key]["Raw"]) // num_channels
                break
        try:
            num_channels_x = num_channels_y = int(np.sqrt(num_channels))
        except NameError:
            raise RuntimeError("No Well found in the file")
        if num_channels_x * num_channels_y != num_channels:
            raise RuntimeError(f"Cannot determine structure of the MEA plate with {num_channels} channels")
        channels = 1 + np.concatenate(np.transpose(np.meshgrid(range(num_channels_x), range(num_channels_y))))

        gain = scale_factor * (max_uv - min_uv) / (max_digital - min_digital)
        offset = min_uv
        read_function = readHDF5t_brw4

        return dict(
            file_handle=rf,
            num_frames=num_frames,
            sampling_rate=sampling_rate,
            num_channels=num_channels,
            channels=channels,
            read_function=read_function,
            gain=gain,
            offset=offset,
        )


def readHDF5t_100(rf, t0, t1, nch):
    return rf["3BData/Raw"][t0:t1]


def readHDF5t_100_i(rf, t0, t1, nch):
    return 4096 - rf["3BData/Raw"][t0:t1]


def readHDF5t_101(rf, t0, t1, nch):
    return rf["3BData/Raw"][nch * t0 : nch * t1].reshape((t1 - t0, nch), order="C")


def readHDF5t_101_i(rf, t0, t1, nch):
    return 4096 - rf["3BData/Raw"][nch * t0 : nch * t1].reshape((t1 - t0, nch), order="C")


def readHDF5t_brw4(rf, t0, t1, nch):
    for key in rf:
        if key[:5] == "Well_":
            return rf[key]["Raw"][nch * t0 : nch * t1].reshape((t1 - t0, nch), order="C")



"""
Class for reading data from pCLAMP and AxoScope
files (.abf version 1 and 2), developed by Molecular device/Axon technologies.

- abf = Axon binary file
- atf is a text file based format from axon that could be
  read by AsciiIO (but this file is less efficient.)


This code is a port of abfload and abf2load
written in Matlab (BSD-2-Clause licence) by :
 - Copyright (c) 2009, Forrest Collman, fcollman@princeton.edu
 - Copyright (c) 2004, Harald Hentschke
and available here:
http://www.mathworks.com/matlabcentral/fileexchange/22114-abf2load


The StringsSection parsing (parse_axon_soup) now relies on an idea
presented in pyABF MIT License Copyright (c) 2018 Scott W Harden
written by Scott Harden. His unofficial documentation for the formats
is here:
https://swharden.com/pyabf/abf2-file-format/
strings section:
[uModifierNameIndex, uCreatorNameIndex, uProtocolPathIndex, lFileComment, lADCCChannelNames, lADCUnitsIndex
lDACChannelNameIndex, lDACUnitIndex, lDACFilePath, nLeakSubtractADC]
['', 'Clampex', '', 'C:/path/protocol.pro', 'some comment', 'IN 0', 'mV', 'IN 1', 'mV', 'Cmd 0', 'pA', 
'Cmd 1', 'pA', 'Cmd 2', 'mV', 'Cmd 3', 'mV']

Information on abf 1 and 2 formats is available here:
http://www.moleculardevices.com/pages/software/developer_info.html

This file supports the old (ABF1) and new (ABF2) format.
ABF1 (clampfit <=9) and ABF2 (clampfit >10)

All possible mode are possible :
    - event-driven variable-length mode 1 -> return several Segments per Block
    - event-driven fixed-length mode 2 or 5 -> return several Segments
    - gap free mode -> return one (or several) Segment in the Block

Supported : Read

Author: Samuel Garcia, JS Nowacki

Note: j.s.nowacki@gmail.com has a C++ library with SWIG bindings which also
reads abf files - would be good to cross-check

"""

import struct
import datetime
from io import open, BufferedReader

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from neo.core import NeoReadWriteError


class AxonRawIO(BaseRawIO):
    """
    Class for Class for reading data from pCLAMP and AxoScope files (.abf version 1 and 2)

    Parameters
    ----------
    filename: str, default: ''
        The *.abf file to be read

    Notes
    -----
    This code is a port of abfload and abf2load written in Matlab (BSD-2-Clause licence) by
    Copyright (c) 2009, Forrest Collman, fcollman@princeton.edu
    Copyright (c) 2004, Harald Hentschke

    Examples
    --------

    >>> import neo.rawio
    >>> reader = neo.rawio.AxonRawIO(filename='mydata.abf')
    >>> reader.parse_header()
    >>> print(reader)

    """

    extensions = ["abf"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _parse_header(self):
        info = self._axon_info = parse_axon_soup(self.filename)

        version = info["fFileVersionNumber"]

        # file format
        if info["nDataFormat"] == 0:
            sig_dtype = np.dtype("i2")
        elif info["nDataFormat"] == 1:
            sig_dtype = np.dtype("f4")

        if version < 2.0:
            nbchannel = info["nADCNumChannels"]
            head_offset = info["lDataSectionPtr"] * BLOCKSIZE + info["nNumPointsIgnored"] * sig_dtype.itemsize
            totalsize = info["lActualAcqLength"]
        elif version >= 2.0:
            nbchannel = info["sections"]["ADCSection"]["llNumEntries"]
            head_offset = info["sections"]["DataSection"]["uBlockIndex"] * BLOCKSIZE
            totalsize = info["sections"]["DataSection"]["llNumEntries"]

        self._raw_data = np.memmap(self.filename, dtype=sig_dtype, mode="r", shape=(totalsize,), offset=head_offset)

        # 3 possible modes
        if version < 2.0:
            mode = info["nOperationMode"]
        elif version >= 2.0:
            mode = info["protocol"]["nOperationMode"]

        if mode not in [1, 2, 3, 5]:
            raise NeoReadWriteError(f"Mode {mode} is not currently supported in Neo")
        # event-driven variable-length mode (mode 1)
        # event-driven fixed-length mode (mode 2 or 5)
        # gap free mode (mode 3) can be in several episodes

        # read sweep pos
        if version < 2.0:
            nbepisod = info["lSynchArraySize"]
            offset_episode = info["lSynchArrayPtr"] * BLOCKSIZE
        elif version >= 2.0:
            nbepisod = info["sections"]["SynchArraySection"]["llNumEntries"]
            offset_episode = info["sections"]["SynchArraySection"]["uBlockIndex"] * BLOCKSIZE
        if nbepisod > 0:
            episode_array = np.memmap(
                self.filename, [("offset", "i4"), ("len", "i4")], "r", shape=nbepisod, offset=offset_episode
            )
        else:
            episode_array = np.empty(1, [("offset", "i4"), ("len", "i4")])
            episode_array[0]["len"] = self._raw_data.size
            episode_array[0]["offset"] = 0

        # sampling_rate
        if version < 2.0:
            self._sampling_rate = 1.0 / (info["fADCSampleInterval"] * nbchannel * 1.0e-6)
        elif version >= 2.0:
            self._sampling_rate = 1.0e6 / info["protocol"]["fADCSequenceInterval"]

        # one sweep = one segment
        nb_segment = episode_array.size

        # Get raw data by segment
        self._raw_signals = {}
        self._t_starts = {}
        pos = 0
        for seg_index in range(nb_segment):
            length = episode_array[seg_index]["len"]

            if version < 2.0:
                fSynchTimeUnit = info["fSynchTimeUnit"]
            elif version >= 2.0:
                fSynchTimeUnit = info["protocol"]["fSynchTimeUnit"]

            if (fSynchTimeUnit != 0) and (mode == 1):
                length /= fSynchTimeUnit

            self._raw_signals[seg_index] = self._raw_data[pos : pos + length].reshape(-1, nbchannel)
            pos += length

            t_start = float(episode_array[seg_index]["offset"])
            if fSynchTimeUnit == 0:
                t_start = t_start / self._sampling_rate
            else:
                t_start = t_start * fSynchTimeUnit * 1e-6
            self._t_starts[seg_index] = t_start

        # Create channel header
        if version < 2.0:
            channel_ids = [chan_num for chan_num in info["nADCSamplingSeq"] if chan_num >= 0]
        else:
            channel_ids = list(range(nbchannel))

        signal_channels = []
        adc_nums = []
        for chan_index, chan_id in enumerate(channel_ids):
            if version < 2.0:
                name = info["sADCChannelName"][chan_id].replace(b" ", b"")
                units = safe_decode_units(info["sADCUnits"][chan_id])
                adc_num = info["nADCPtoLChannelMap"][chan_id]
            elif version >= 2.0:
                ADCInfo = info["listADCInfo"][chan_id]
                name = ADCInfo["ADCChNames"].replace(b" ", b"")
                units = safe_decode_units(ADCInfo["ADCChUnits"])
                adc_num = ADCInfo["nADCNum"]
            adc_nums.append(adc_num)

            if info["nDataFormat"] == 0:
                # int16 gain/offset
                if version < 2.0:
                    gain = info["fADCRange"]
                    gain /= info["fInstrumentScaleFactor"][chan_id]
                    gain /= info["fSignalGain"][chan_id]
                    gain /= info["fADCProgrammableGain"][chan_id]
                    gain /= info["lADCResolution"]
                    if info["nTelegraphEnable"][chan_id] == 0:
                        pass
                    elif info["nTelegraphEnable"][chan_id] == 1:
                        gain /= info["fTelegraphAdditGain"][chan_id]
                    else:
                        self.logger.warning("ignoring buggy nTelegraphEnable")
                    offset = info["fInstrumentOffset"][chan_id]
                    offset -= info["fSignalOffset"][chan_id]
                elif version >= 2.0:
                    gain = info["protocol"]["fADCRange"]
                    gain /= info["listADCInfo"][chan_id]["fInstrumentScaleFactor"]
                    gain /= info["listADCInfo"][chan_id]["fSignalGain"]
                    gain /= info["listADCInfo"][chan_id]["fADCProgrammableGain"]
                    gain /= info["protocol"]["lADCResolution"]
                    if info["listADCInfo"][chan_id]["nTelegraphEnable"]:
                        gain /= info["listADCInfo"][chan_id]["fTelegraphAdditGain"]
                    offset = info["listADCInfo"][chan_id]["fInstrumentOffset"]
                    offset -= info["listADCInfo"][chan_id]["fSignalOffset"]
            else:
                gain, offset = 1.0, 0.0
            stream_id = "0"
            signal_channels.append((name, str(chan_id), self._sampling_rate, sig_dtype, units, gain, offset, stream_id))

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # one unique signal stream
        signal_streams = np.array([("Signals", "0")], dtype=_signal_stream_dtype)

        # only one events channel : tag
        # In ABF timstamps are not attached too any particular segment
        # so each segment acess all event
        timestamps = []
        labels = []
        comments = []
        for i, tag in enumerate(info["listTag"]):
            timestamps.append(tag["lTagTime"])
            labels.append(str(tag["nTagType"]))
            comments.append(str(clean_string(tag["sComment"])))
        self._raw_ev_timestamps = np.array(timestamps)
        self._ev_labels = np.array(labels, dtype="U")
        self._ev_comments = np.array(comments, dtype="U")
        event_channels = [("Tag", "", "event")]
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()
        bl_annotations = self.raw_annotations["blocks"][0]

        bl_annotations["rec_datetime"] = info["rec_datetime"]
        bl_annotations["abf_version"] = version

        for seg_index in range(nb_segment):
            seg_annotations = bl_annotations["segments"][seg_index]
            seg_annotations["abf_version"] = version

            signal_an = self.raw_annotations["blocks"][0]["segments"][seg_index]["signals"][0]
            nADCNum = np.array([adc_nums[c] for c in range(signal_channels.size)])
            signal_an["__array_annotations__"]["nADCNum"] = nADCNum

            for c in range(event_channels.size):
                ev_ann = seg_annotations["events"][c]
                ev_ann["comments"] = self._ev_comments

    def _source_name(self):
        return self.filename

    def _segment_t_start(self, block_index, seg_index):
        return self._t_starts[seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._t_starts[seg_index] + self._raw_signals[seg_index].shape[0] / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        shape = self._raw_signals[seg_index].shape
        return shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return self._t_starts[seg_index]

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(None)
        raw_signals = self._raw_signals[seg_index][slice(i_start, i_stop), channel_indexes]
        return raw_signals

    def _event_count(self, block_index, seg_index, event_channel_index):
        return self._raw_ev_timestamps.size

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        # In ABF timestamps are not attached too any particular segment
        # so each segment accesses all events
        timestamp = self._raw_ev_timestamps
        labels = self._ev_labels
        durations = None

        if t_start is not None:
            keep = timestamp >= int(t_start * self._sampling_rate)
            timestamp = timestamp[keep]
            labels = labels[keep]

        if t_stop is not None:
            keep = timestamp <= int(t_stop * self._sampling_rate)
            timestamp = timestamp[keep]
            labels = labels[keep]

        return timestamp, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype) / self._sampling_rate
        return event_times

    def read_raw_protocol(self):
        """
        Read the protocol waveform of the file, if present;
        function works with ABF2 only. Protocols can be reconstructed
        from the ABF1 header.

        Returns
        -------
        segments : list of segments
            Segments, one for every episode, with list of analog signls (one for every DAC).

        Author:  JS Nowacki
        """
        info = self._axon_info

        if info["fFileVersionNumber"] < 2.0:
            raise IOError("Protocol section is only present in ABF2 files.")

        nADC = info["sections"]["ADCSection"]["llNumEntries"]  # Number of ADC channels
        nDAC = info["sections"]["DACSection"]["llNumEntries"]  # Number of DAC channels
        nSam = int(info["protocol"]["lNumSamplesPerEpisode"] / nADC)  # Number of samples per episode
        nEpi = info["lActualEpisodes"]  # Actual number of episodes

        # Make a list of segments with analog signals with just holding levels
        # List of segments relates to number of episodes, as for recorded data
        sigs_by_segments = []
        for epiNum in range(nEpi):
            # One analog signal for each DAC in segment (episode)
            signals = []
            for DACNum in range(nDAC):
                sig = np.ones(nSam) * info["listDACInfo"][DACNum]["fDACHoldingLevel"]
                # If there are epoch infos for this DAC
                if DACNum in info["dictEpochInfoPerDAC"]:
                    # Save last sample index
                    i_last = int(nSam * 15625 / 10**6)
                    # TODO guess for first holding
                    # Go over EpochInfoPerDAC and change the analog signal
                    # according to the epochs
                    epochInfo = info["dictEpochInfoPerDAC"][DACNum]
                    for epochNum, epoch in epochInfo.items():
                        i_begin = i_last
                        i_end = i_last + epoch["lEpochInitDuration"] + epoch["lEpochDurationInc"] * epiNum
                        dif = i_end - i_begin
                        sig[i_begin:i_end] = np.ones(dif) * (
                            epoch["fEpochInitLevel"] + epoch["fEpochLevelInc"] * epiNum
                        )
                        i_last += epoch["lEpochInitDuration"] + epoch["lEpochDurationInc"] * epiNum
                signals.append(sig)
            sigs_by_segments.append(signals)

        sig_names = []
        sig_units = []
        for DACNum in range(nDAC):
            name = info["listDACInfo"][DACNum]["DACChNames"].decode("utf-8")
            units = safe_decode_units(info["listDACInfo"][DACNum]["DACChUnits"])
            sig_names.append(name)
            sig_units.append(units)

        return sigs_by_segments, sig_names, sig_units


def parse_axon_soup(filename):
    """
    read the header of the file

    The strategy here differs from the original script under Matlab.
    In the original script for ABF2, it completes the header with
    information that is located in other structures.

    In ABF2 this function returns info with sub dict:
        sections             (ABF2)
        protocol             (ABF2)
        listTags             (ABF1&2)
        listADCInfo          (ABF2)
        listDACInfo          (ABF2)
        dictEpochInfoPerDAC  (ABF2)
    that contains more information.
    """
    with open(filename, "rb") as fid:
        f = StructFile(fid)

        # version
        f_file_signature = f.read(4)
        if f_file_signature == b"ABF ":
            header_description = headerDescriptionV1
        elif f_file_signature == b"ABF2":
            header_description = headerDescriptionV2
        else:
            return None

        # construct dict
        header = {}
        for key, offset, fmt in header_description:
            val = f.read_f(fmt, offset=offset)
            if len(val) == 1:
                header[key] = val[0]
            else:
                header[key] = np.array(val)

        # correction of version number and starttime
        if f_file_signature == b"ABF ":
            header["lFileStartTime"] += header["nFileStartMillisecs"] * 0.001
        elif f_file_signature == b"ABF2":
            n = header["fFileVersionNumber"]
            header["fFileVersionNumber"] = n[3] + 0.1 * n[2] + 0.01 * n[1] + 0.001 * n[0]
            header["lFileStartTime"] = header["uFileStartTimeMS"] * 0.001

        if header["fFileVersionNumber"] < 2.0:
            # tags
            listTag = []
            for i in range(header["lNumTagEntries"]):
                f.seek(header["lTagSectionPtr"] + i * 64)
                tag = {}
                for key, fmt in TagInfoDescription:
                    val = f.read_f(fmt)
                    if len(val) == 1:
                        tag[key] = val[0]
                    else:
                        tag[key] = np.array(val)
                listTag.append(tag)
            header["listTag"] = listTag
            # protocol name formatting
            header["sProtocolPath"] = clean_string(header["sProtocolPath"])
            header["sProtocolPath"] = header["sProtocolPath"].replace(b"\\", b"/")

        elif header["fFileVersionNumber"] >= 2.0:
            # in abf2 some info are in other place

            # sections
            sections = {}
            for s, sectionName in enumerate(sectionNames):
                uBlockIndex, uBytes, llNumEntries = f.read_f("IIl", offset=76 + s * 16)
                sections[sectionName] = {}
                sections[sectionName]["uBlockIndex"] = uBlockIndex
                sections[sectionName]["uBytes"] = uBytes
                sections[sectionName]["llNumEntries"] = llNumEntries
            header["sections"] = sections

            # strings sections
            # hack for reading channels names and units
            # this section is not very detailed and so the code
            # not very robust.
            f.seek(sections["StringsSection"]["uBlockIndex"] * BLOCKSIZE)
            big_string = f.read(sections["StringsSection"]["uBytes"])
            # this idea comes from pyABF https://github.com/swharden/pyABF
            # previously we searched for clampex, Clampex etc, but this was
            # brittle. pyABF believes that looking for the \x00\x00 is more
            # robust. We find these values, replace mu->u, then split into
            # a set of strings
            indexed_string = big_string[big_string.rfind(b"\x00\x00") :]
            # replace mu -> u for easy display
            indexed_string = indexed_string.replace(b"\xb5", b"\x75")
            # we need to remove one of the \x00 to have the indices be
            # the correct order
            indexed_string = indexed_string.split(b"\x00")[1:]
            strings = indexed_string

            # ADC sections
            header["listADCInfo"] = []
            for i in range(sections["ADCSection"]["llNumEntries"]):
                # read ADCInfo
                f.seek(sections["ADCSection"]["uBlockIndex"] * BLOCKSIZE + sections["ADCSection"]["uBytes"] * i)
                ADCInfo = {}
                for key, fmt in ADCInfoDescription:
                    val = f.read_f(fmt)
                    if len(val) == 1:
                        ADCInfo[key] = val[0]
                    else:
                        ADCInfo[key] = np.array(val)
                ADCInfo["ADCChNames"] = strings[ADCInfo["lADCChannelNameIndex"]]
                ADCInfo["ADCChUnits"] = strings[ADCInfo["lADCUnitsIndex"]]
                header["listADCInfo"].append(ADCInfo)

            # protocol sections
            protocol = {}
            f.seek(sections["ProtocolSection"]["uBlockIndex"] * BLOCKSIZE)
            for key, fmt in protocolInfoDescription:
                val = f.read_f(fmt)
                if len(val) == 1:
                    protocol[key] = val[0]
                else:
                    protocol[key] = np.array(val)
            header["protocol"] = protocol
            header["sProtocolPath"] = strings[header["uProtocolPathIndex"]]

            # tags
            listTag = []
            for i in range(sections["TagSection"]["llNumEntries"]):
                f.seek(sections["TagSection"]["uBlockIndex"] * BLOCKSIZE + sections["TagSection"]["uBytes"] * i)
                tag = {}
                for key, fmt in TagInfoDescription:
                    val = f.read_f(fmt)
                    if len(val) == 1:
                        tag[key] = val[0]
                    else:
                        tag[key] = np.array(val)
                listTag.append(tag)

            header["listTag"] = listTag

            # DAC sections
            header["listDACInfo"] = []
            for i in range(sections["DACSection"]["llNumEntries"]):
                # read DACInfo
                f.seek(sections["DACSection"]["uBlockIndex"] * BLOCKSIZE + sections["DACSection"]["uBytes"] * i)
                DACInfo = {}
                for key, fmt in DACInfoDescription:
                    val = f.read_f(fmt)
                    if len(val) == 1:
                        DACInfo[key] = val[0]
                    else:
                        DACInfo[key] = np.array(val)
                DACInfo["DACChNames"] = strings[DACInfo["lDACChannelNameIndex"]]
                DACInfo["DACChUnits"] = strings[DACInfo["lDACChannelUnitsIndex"]]

                header["listDACInfo"].append(DACInfo)

            # EpochPerDAC  sections
            # header['dictEpochInfoPerDAC'] is dict of dicts:
            #  - the first index is the DAC number
            #  - the second index is the epoch number
            # It has to be done like that because data may not exist
            # and may not be in sorted order
            header["dictEpochInfoPerDAC"] = {}
            for i in range(sections["EpochPerDACSection"]["llNumEntries"]):
                #  read DACInfo
                f.seek(
                    sections["EpochPerDACSection"]["uBlockIndex"] * BLOCKSIZE
                    + sections["EpochPerDACSection"]["uBytes"] * i
                )
                EpochInfoPerDAC = {}
                for key, fmt in EpochInfoPerDACDescription:
                    val = f.read_f(fmt)
                    if len(val) == 1:
                        EpochInfoPerDAC[key] = val[0]
                    else:
                        EpochInfoPerDAC[key] = np.array(val)

                DACNum = EpochInfoPerDAC["nDACNum"]
                EpochNum = EpochInfoPerDAC["nEpochNum"]
                # Checking if the key exists, if not, the value is empty
                # so we have to create empty dict to populate
                if DACNum not in header["dictEpochInfoPerDAC"]:
                    header["dictEpochInfoPerDAC"][DACNum] = {}

                header["dictEpochInfoPerDAC"][DACNum][EpochNum] = EpochInfoPerDAC

            # Epoch sections
            header["EpochInfo"] = []
            for i in range(sections["EpochSection"]["llNumEntries"]):
                # read EpochInfo
                f.seek(sections["EpochSection"]["uBlockIndex"] * BLOCKSIZE + sections["EpochSection"]["uBytes"] * i)
                EpochInfo = {}
                for key, fmt in EpochInfoDescription:
                    val = f.read_f(fmt)
                    if len(val) == 1:
                        EpochInfo[key] = val[0]
                    else:
                        EpochInfo[key] = np.array(val)
                header["EpochInfo"].append(EpochInfo)

        # date and time
        if header["fFileVersionNumber"] < 2.0:
            YY = 1900
            MM = 1
            DD = 1
            hh = int(header["lFileStartTime"] / 3600.0)
            mm = int((header["lFileStartTime"] - hh * 3600) / 60)
            ss = header["lFileStartTime"] - hh * 3600 - mm * 60
            ms = int(np.mod(ss, 1) * 1e6)
            ss = int(ss)
        elif header["fFileVersionNumber"] >= 2.0:
            YY = int(header["uFileStartDate"] / 10000)
            MM = int((header["uFileStartDate"] - YY * 10000) / 100)
            DD = int(header["uFileStartDate"] - YY * 10000 - MM * 100)
            hh = int(header["uFileStartTimeMS"] / 1000.0 / 3600.0)
            mm = int((header["uFileStartTimeMS"] / 1000.0 - hh * 3600) / 60)
            ss = header["uFileStartTimeMS"] / 1000.0 - hh * 3600 - mm * 60
            ms = int(np.mod(ss, 1) * 1e6)
            ss = int(ss)
        header["rec_datetime"] = datetime.datetime(YY, MM, DD, hh, mm, ss, ms)

    return header


class StructFile(BufferedReader):
    def read_f(self, fmt, offset=None):
        if offset is not None:
            self.seek(offset)
        return struct.unpack(fmt, self.read(struct.calcsize(fmt)))


def clean_string(s):
    s = s.rstrip(b"\x00")
    s = s.rstrip(b" ")
    return s


def safe_decode_units(s):
    s = s.replace(b" ", b"")
    s = s.replace(b"\xb5", b"u")  # \xb5 is µ
    s = s.replace(b"\xb0", b"\xc2\xb0")  # \xb0 is °
    s = s.decode("utf-8")
    return s


BLOCKSIZE = 512

headerDescriptionV1 = [
    ("fFileSignature", 0, "4s"),
    ("fFileVersionNumber", 4, "f"),
    ("nOperationMode", 8, "h"),
    ("lActualAcqLength", 10, "i"),
    ("nNumPointsIgnored", 14, "h"),
    ("lActualEpisodes", 16, "i"),
    ("lFileStartTime", 24, "i"),
    ("lDataSectionPtr", 40, "i"),
    ("lTagSectionPtr", 44, "i"),
    ("lNumTagEntries", 48, "i"),
    ("lSynchArrayPtr", 92, "i"),
    ("lSynchArraySize", 96, "i"),
    ("nDataFormat", 100, "h"),
    ("nADCNumChannels", 120, "h"),
    ("fADCSampleInterval", 122, "f"),
    ("fSynchTimeUnit", 130, "f"),
    ("lNumSamplesPerEpisode", 138, "i"),
    ("lPreTriggerSamples", 142, "i"),
    ("lEpisodesPerRun", 146, "i"),
    ("fADCRange", 244, "f"),
    ("lADCResolution", 252, "i"),
    ("nFileStartMillisecs", 366, "h"),
    ("nADCPtoLChannelMap", 378, "16h"),
    ("nADCSamplingSeq", 410, "16h"),
    ("sADCChannelName", 442, "10s" * 16),
    ("sADCUnits", 602, "8s" * 16),
    ("fADCProgrammableGain", 730, "16f"),
    ("fInstrumentScaleFactor", 922, "16f"),
    ("fInstrumentOffset", 986, "16f"),
    ("fSignalGain", 1050, "16f"),
    ("fSignalOffset", 1114, "16f"),
    ("nDigitalEnable", 1436, "h"),
    ("nActiveDACChannel", 1440, "h"),
    ("nDigitalHolding", 1584, "h"),
    ("nDigitalInterEpisode", 1586, "h"),
    ("nDigitalValue", 2588, "10h"),
    ("lDACFilePtr", 2048, "2i"),
    ("lDACFileNumEpisodes", 2056, "2i"),
    ("fDACCalibrationFactor", 2074, "4f"),
    ("fDACCalibrationOffset", 2090, "4f"),
    ("nWaveformEnable", 2296, "2h"),
    ("nWaveformSource", 2300, "2h"),
    ("nInterEpisodeLevel", 2304, "2h"),
    ("nEpochType", 2308, "20h"),
    ("fEpochInitLevel", 2348, "20f"),
    ("fEpochLevelInc", 2428, "20f"),
    ("lEpochInitDuration", 2508, "20i"),
    ("lEpochDurationInc", 2588, "20i"),
    ("nTelegraphEnable", 4512, "16h"),
    ("fTelegraphAdditGain", 4576, "16f"),
    ("sProtocolPath", 4898, "384s"),
]

headerDescriptionV2 = [
    ("fFileSignature", 0, "4s"),
    ("fFileVersionNumber", 4, "4b"),
    ("uFileInfoSize", 8, "I"),
    ("lActualEpisodes", 12, "I"),
    ("uFileStartDate", 16, "I"),
    ("uFileStartTimeMS", 20, "I"),
    ("uStopwatchTime", 24, "I"),
    ("nFileType", 28, "H"),
    ("nDataFormat", 30, "H"),
    ("nSimultaneousScan", 32, "H"),
    ("nCRCEnable", 34, "H"),
    ("uFileCRC", 36, "I"),
    ("FileGUID", 40, "I"),
    ("uCreatorVersion", 56, "I"),
    ("uCreatorNameIndex", 60, "I"),
    ("uModifierVersion", 64, "I"),
    ("uModifierNameIndex", 68, "I"),
    ("uProtocolPathIndex", 72, "I"),
]

sectionNames = [
    "ProtocolSection",
    "ADCSection",
    "DACSection",
    "EpochSection",
    "ADCPerDACSection",
    "EpochPerDACSection",
    "UserListSection",
    "StatsRegionSection",
    "MathSection",
    "StringsSection",
    "DataSection",
    "TagSection",
    "ScopeSection",
    "DeltaSection",
    "VoiceTagSection",
    "SynchArraySection",
    "AnnotationSection",
    "StatsSection",
]

protocolInfoDescription = [
    ("nOperationMode", "h"),
    ("fADCSequenceInterval", "f"),
    ("bEnableFileCompression", "b"),
    ("sUnused1", "3s"),
    ("uFileCompressionRatio", "I"),
    ("fSynchTimeUnit", "f"),
    ("fSecondsPerRun", "f"),
    ("lNumSamplesPerEpisode", "i"),
    ("lPreTriggerSamples", "i"),
    ("lEpisodesPerRun", "i"),
    ("lRunsPerTrial", "i"),
    ("lNumberOfTrials", "i"),
    ("nAveragingMode", "h"),
    ("nUndoRunCount", "h"),
    ("nFirstEpisodeInRun", "h"),
    ("fTriggerThreshold", "f"),
    ("nTriggerSource", "h"),
    ("nTriggerAction", "h"),
    ("nTriggerPolarity", "h"),
    ("fScopeOutputInterval", "f"),
    ("fEpisodeStartToStart", "f"),
    ("fRunStartToStart", "f"),
    ("lAverageCount", "i"),
    ("fTrialStartToStart", "f"),
    ("nAutoTriggerStrategy", "h"),
    ("fFirstRunDelayS", "f"),
    ("nChannelStatsStrategy", "h"),
    ("lSamplesPerTrace", "i"),
    ("lStartDisplayNum", "i"),
    ("lFinishDisplayNum", "i"),
    ("nShowPNRawData", "h"),
    ("fStatisticsPeriod", "f"),
    ("lStatisticsMeasurements", "i"),
    ("nStatisticsSaveStrategy", "h"),
    ("fADCRange", "f"),
    ("fDACRange", "f"),
    ("lADCResolution", "i"),
    ("lDACResolution", "i"),
    ("nExperimentType", "h"),
    ("nManualInfoStrategy", "h"),
    ("nCommentsEnable", "h"),
    ("lFileCommentIndex", "i"),
    ("nAutoAnalyseEnable", "h"),
    ("nSignalType", "h"),
    ("nDigitalEnable", "h"),
    ("nActiveDACChannel", "h"),
    ("nDigitalHolding", "h"),
    ("nDigitalInterEpisode", "h"),
    ("nDigitalDACChannel", "h"),
    ("nDigitalTrainActiveLogic", "h"),
    ("nStatsEnable", "h"),
    ("nStatisticsClearStrategy", "h"),
    ("nLevelHysteresis", "h"),
    ("lTimeHysteresis", "i"),
    ("nAllowExternalTags", "h"),
    ("nAverageAlgorithm", "h"),
    ("fAverageWeighting", "f"),
    ("nUndoPromptStrategy", "h"),
    ("nTrialTriggerSource", "h"),
    ("nStatisticsDisplayStrategy", "h"),
    ("nExternalTagType", "h"),
    ("nScopeTriggerOut", "h"),
    ("nLTPType", "h"),
    ("nAlternateDACOutputState", "h"),
    ("nAlternateDigitalOutputState", "h"),
    ("fCellID", "3f"),
    ("nDigitizerADCs", "h"),
    ("nDigitizerDACs", "h"),
    ("nDigitizerTotalDigitalOuts", "h"),
    ("nDigitizerSynchDigitalOuts", "h"),
    ("nDigitizerType", "h"),
]

ADCInfoDescription = [
    ("nADCNum", "h"),
    ("nTelegraphEnable", "h"),
    ("nTelegraphInstrument", "h"),
    ("fTelegraphAdditGain", "f"),
    ("fTelegraphFilter", "f"),
    ("fTelegraphMembraneCap", "f"),
    ("nTelegraphMode", "h"),
    ("fTelegraphAccessResistance", "f"),
    ("nADCPtoLChannelMap", "h"),
    ("nADCSamplingSeq", "h"),
    ("fADCProgrammableGain", "f"),
    ("fADCDisplayAmplification", "f"),
    ("fADCDisplayOffset", "f"),
    ("fInstrumentScaleFactor", "f"),
    ("fInstrumentOffset", "f"),
    ("fSignalGain", "f"),
    ("fSignalOffset", "f"),
    ("fSignalLowpassFilter", "f"),
    ("fSignalHighpassFilter", "f"),
    ("nLowpassFilterType", "b"),
    ("nHighpassFilterType", "b"),
    ("fPostProcessLowpassFilter", "f"),
    ("nPostProcessLowpassFilterType", "c"),
    ("bEnabledDuringPN", "b"),
    ("nStatsChannelPolarity", "h"),
    ("lADCChannelNameIndex", "i"),
    ("lADCUnitsIndex", "i"),
]

TagInfoDescription = [
    ("lTagTime", "i"),
    ("sComment", "56s"),
    ("nTagType", "h"),
    ("nVoiceTagNumber_or_AnnotationIndex", "h"),
]

DACInfoDescription = [
    ("nDACNum", "h"),
    ("nTelegraphDACScaleFactorEnable", "h"),
    ("fInstrumentHoldingLevel", "f"),
    ("fDACScaleFactor", "f"),
    ("fDACHoldingLevel", "f"),
    ("fDACCalibrationFactor", "f"),
    ("fDACCalibrationOffset", "f"),
    ("lDACChannelNameIndex", "i"),
    ("lDACChannelUnitsIndex", "i"),
    ("lDACFilePtr", "i"),
    ("lDACFileNumEpisodes", "i"),
    ("nWaveformEnable", "h"),
    ("nWaveformSource", "h"),
    ("nInterEpisodeLevel", "h"),
    ("fDACFileScale", "f"),
    ("fDACFileOffset", "f"),
    ("lDACFileEpisodeNum", "i"),
    ("nDACFileADCNum", "h"),
    ("nConditEnable", "h"),
    ("lConditNumPulses", "i"),
    ("fBaselineDuration", "f"),
    ("fBaselineLevel", "f"),
    ("fStepDuration", "f"),
    ("fStepLevel", "f"),
    ("fPostTrainPeriod", "f"),
    ("fPostTrainLevel", "f"),
    ("nMembTestEnable", "h"),
    ("nLeakSubtractType", "h"),
    ("nPNPolarity", "h"),
    ("fPNHoldingLevel", "f"),
    ("nPNNumADCChannels", "h"),
    ("nPNPosition", "h"),
    ("nPNNumPulses", "h"),
    ("fPNSettlingTime", "f"),
    ("fPNInterpulse", "f"),
    ("nLTPUsageOfDAC", "h"),
    ("nLTPPresynapticPulses", "h"),
    ("lDACFilePathIndex", "i"),
    ("fMembTestPreSettlingTimeMS", "f"),
    ("fMembTestPostSettlingTimeMS", "f"),
    ("nLeakSubtractADCIndex", "h"),
    ("sUnused", "124s"),
]

EpochInfoPerDACDescription = [
    ("nEpochNum", "h"),
    ("nDACNum", "h"),
    ("nEpochType", "h"),
    ("fEpochInitLevel", "f"),
    ("fEpochLevelInc", "f"),
    ("lEpochInitDuration", "i"),
    ("lEpochDurationInc", "i"),
    ("lEpochPulsePeriod", "i"),
    ("lEpochPulseWidth", "i"),
    ("sUnused", "18s"),
]

EpochInfoDescription = [
    ("nEpochNum", "h"),
    ("nDigitalValue", "h"),
    ("nDigitalTrainValue", "h"),
    ("nAlternateDigitalValue", "h"),
    ("nAlternateDigitalTrainValue", "h"),
    ("bEpochCompression", "b"),
    ("sUnused", "21s"),
]



"""
This module implements :class:`ChannelView`, which represents a subset of the
channels in an :class:`AnalogSignal` or :class:`IrregularlySampledSignal`.

It replaces the indexing function of the former :class:`ChannelIndex`.
"""

import numpy as np

from .baseneo import BaseNeo
from .basesignal import BaseSignal
from .dataobject import ArrayDict


class ChannelView(BaseNeo):
    """
    A tool for indexing a subset of the channels within an :class:`AnalogSignal`
    or :class:`IrregularlySampledSignal`;

    Parameters
    ----------
    obj: Neo.AnalogSignal | Neo.IrregularlySampledSignal
        The neo object to index
    index: list | np.ndarray
        Boolean or integer mask to select the channels of interest
    name: str | None, default: None
        A label for the dataset.
    description: str | None, default: None
         Text description.
    file_origin: str | None, default: None
        Filesystem path or URL of the original data file.
    array_annotations: dict | None, default: None
        Dict mapping strings to numpy arrays containing annotations for all data points
    **annotations: dict
        Other use-specified metadata with keywords
    """

    _parent_objects = ("Group",)
    _parent_attrs = ("group",)
    _necessary_attrs = (
        ("obj", ("AnalogSignal", "IrregularlySampledSignal"), 1),
        ("index", np.ndarray, 1, np.dtype("i")),
    )
    is_view = True

    # "mask" would be an alternative name, proposing "index" for
    # backwards-compatibility with ChannelIndex

    def __init__(
        self, obj, index, name=None, description=None, file_origin=None, array_annotations=None, **annotations
    ):
        super().__init__(name=name, description=description, file_origin=file_origin, **annotations)

        if not (isinstance(obj, BaseSignal) or (hasattr(obj, "proxy_for") and issubclass(obj.proxy_for, BaseSignal))):
            raise ValueError("Can only take a ChannelView of an AnalogSignal " "or an IrregularlySampledSignal")
        self.obj = obj

        # check type and dtype of index and convert index to a common form
        # (accept list or array of bool or int, convert to int array)
        self.index = np.array(index)
        if len(self.index.shape) != 1:
            raise ValueError("index must be a 1D array")
        if self.index.dtype == bool:  # convert boolean mask to integer index
            if self.index.size != self.obj.shape[-1]:
                raise ValueError("index size does not match number of channels in signal")
            (self.index,) = np.nonzero(self.index)
        # allow any type of integer representation
        elif self.index.dtype.char not in np.typecodes["AllInteger"]:
            raise ValueError("index must be of a list or array of data type boolean or integer")

        if not hasattr(self, "array_annotations") or not self.array_annotations:
            self.array_annotations = ArrayDict(self._get_arr_ann_length())
        if array_annotations is not None:
            self.array_annotate(**array_annotations)

    @property
    def shape(self):
        return (self.obj.shape[0], self.index.size)

    def _get_arr_ann_length(self):
        return self.index.size

    def array_annotate(self, **array_annotations):
        self.array_annotations.update(array_annotations)

    def resolve(self):
        """
        Return a copy of the underlying object containing just the subset of channels
        defined by the index.
        """
        return self.obj[:, self.index]



import quantities as pq
from neo.core.regionofinterest import RectangularRegionOfInterest, CircularRegionOfInterest, PolygonRegionOfInterest
from neo.core.imagesequence import ImageSequence
import unittest


class Test_CircularRegionOfInterest(unittest.TestCase):

    def test_result(self):
        seq = ImageSequence([[[]]], spatial_scale=1, frame_duration=20 * pq.ms)
        self.assertEqual((CircularRegionOfInterest(seq, 6, 6, 1).pixels_in_region()), [[6, 5], [5, 6], [6, 6]])
        self.assertEqual(
            (CircularRegionOfInterest(seq, 6, 6, 1.01).pixels_in_region()), [[6, 5], [5, 6], [6, 6], [7, 6], [6, 7]]
        )


class Test_RectangularRegionOfInterest(unittest.TestCase):

    def test_result(self):
        seq = ImageSequence([[[]]], spatial_scale=1, frame_duration=20 * pq.ms)
        self.assertEqual(
            RectangularRegionOfInterest(seq, 5, 5, 2, 2).pixels_in_region(), [[4, 4], [5, 4], [4, 5], [5, 5]]
        )


class Test_PolygonRegionOfInterest(unittest.TestCase):

    def test_result(self):
        seq = ImageSequence([[[]]], spatial_scale=1, frame_duration=20 * pq.ms)
        self.assertEqual(
            PolygonRegionOfInterest(seq, (3, 3), (2, 5), (5, 5), (5, 1), (1, 1)).pixels_in_region(),
            [(1, 1), (2, 1), (3, 1), (4, 1), (2, 2), (3, 2), (4, 2), (3, 3), (4, 3), (3, 4), (4, 4)],
        )


if __name__ == "__main__":
    unittest.main()



"""
Class for reading data from maxwell biosystem device:
  * MaxOne
  * MaxTwo

https://www.mxwbio.com/resources/mea/

The implementation is a mix between:
  * the implementation in spikeextractors
     https://github.com/SpikeInterface/spikeextractors/blob/master/spikeextractors/extractors/maxwellextractors/maxwellextractors.py
 * the implementation in spyking-circus
    https://github.com/spyking-circus/spyking-circus/blob/master/circus/files/maxwell.py

This implementation does not handle spikes at the moment.

For maxtwo device, each well will be a different signal stream.

Author : Samuel Garcia, Alessio Buccino, Pierre Yger
"""

import os
from pathlib import Path
import platform
from urllib.request import urlopen

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

from neo.core import NeoReadWriteError


class MaxwellRawIO(BaseRawIO):
    """
    Class for reading MaxOne or MaxTwo files.

    Parameters
    ----------

    filename: str, default: ''
        The *.h5 file to be loaded
    rec_name: str | None, default: None
        If the file has multiple recordings, specify the one to read.
        For 24-well plates, the rec_name needs to be specified since different well
        rows generate different recording ids.
        E.g., rec0001, rec0002, etc.

    """

    extensions = ["h5"]
    rawmode = "one-file"

    def __init__(self, filename="", rec_name=None):
        BaseRawIO.__init__(self)
        self.filename = filename
        self.rec_name = rec_name

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        import h5py

        h5file = h5py.File(self.filename, mode="r")
        self.h5_file = h5file
        version = h5file["version"][0].decode()

        # create signal stream
        # one stream per well
        signal_streams = []
        if int(version) == 20160704:
            self._old_format = True
            signal_streams.append(("well000", "well000"))
        elif int(version) > 20160704:
            # multi stream stream (one well is one stream)
            self._old_format = False
            well_ids = list(h5file["wells"].keys())
            unique_rec_names = []
            for well_name in well_ids:
                rec_names = list(h5file["wells"][well_name].keys())
                for rec_name in rec_names:
                    unique_rec_names.append(rec_name)
            # check consistency of rec_names
            unique_rec_names = np.unique(unique_rec_names)
            if len(unique_rec_names) > 1:
                if self.rec_name is None:
                    raise ValueError(
                        f"Detected multiple recording IDs across wells. "
                        f"Please select a single recording using the `rec_name` parameter. "
                        f"Possible rec_names: {unique_rec_names}"
                    )
                else:
                    if self.rec_name not in unique_rec_names:
                        raise NeoReadWriteError(f"rec_name {self.rec_name} not found")
            else:
                self.rec_name = unique_rec_names[0]
            # add streams that contain the selected rec_name
            for well_name in well_ids:
                rec_names = list(h5file["wells"][well_name].keys())
                if self.rec_name in rec_names:
                    signal_streams.append((well_name, well_name))
        else:
            raise NotImplementedError(f"This version {version} is not supported")

        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # create signal channels
        max_sig_length = 0
        self._signals = {}
        sig_channels = []
        for stream_id in signal_streams["id"]:
            if int(version) == 20160704:
                sr = 20000.0
                settings = h5file["settings"]
                if "lsb" in settings:
                    gain_uV = settings["lsb"][0] * 1e6
                else:
                    if "gain" not in settings:
                        print("'gain' amd 'lsb' not found in settings. " "Setting gain to 512 (default)")
                        gain = 512
                    else:
                        gain = settings["gain"][0]
                    gain_uV = 3.3 / (1024 * gain) * 1e6
                sigs = h5file["sig"]
                mapping = h5file["mapping"]
                ids = np.array(mapping["channel"])
                ids = ids[ids >= 0]
                self._channel_slice = ids
            elif int(version) > 20160704:
                settings = h5file["wells"][stream_id][self.rec_name]["settings"]
                sr = settings["sampling"][0]
                if "lsb" in settings:
                    gain_uV = settings["lsb"][0] * 1e6
                else:
                    if "gain" not in settings:
                        print("'gain' amd 'lsb' not found in settings. " "Setting gain to 512 (default)")
                        gain = 512
                    else:
                        gain = settings["gain"][0]
                    gain_uV = 3.3 / (1024 * gain) * 1e6
                mapping = settings["mapping"]
                sigs = h5file["wells"][stream_id][self.rec_name]["groups"]["routed"]["raw"]

            channel_ids = np.array(mapping["channel"])
            electrode_ids = np.array(mapping["electrode"])
            mask = channel_ids >= 0
            channel_ids = channel_ids[mask]
            electrode_ids = electrode_ids[mask]

            for i, chan_id in enumerate(channel_ids):
                elec_id = electrode_ids[i]
                ch_name = f"ch{chan_id} elec{elec_id}"
                offset_uV = 0
                sig_channels.append((ch_name, str(chan_id), sr, "uint16", "uV", gain_uV, offset_uV, stream_id))

            self._signals[stream_id] = sigs
            max_sig_length = max(max_sig_length, sigs.shape[1])

        self._t_stop = max_sig_length / sr

        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()
        bl_ann = self.raw_annotations["blocks"][0]
        bl_ann["maxwell_version"] = version

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        return self._t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        sigs = self._signals[stream_id]
        return sigs.shape[1]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        sigs = self._signals[stream_id]

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = sigs.shape[1]

        resorted_indexes = None
        if channel_indexes is None:
            channel_indexes = slice(None)
        else:
            if np.array(channel_indexes).size > 1 and np.any(np.diff(channel_indexes) < 0):
                # get around h5py constraint that it does not allow datasets
                # to be indexed out of order
                order_f = np.argsort(channel_indexes)
                sorted_channel_indexes = channel_indexes[order_f]
                # use argsort again on order_f to obtain resorted_indexes
                resorted_indexes = np.argsort(order_f)

        try:
            if resorted_indexes is None:
                if self._old_format:
                    sigs = sigs[self._channel_slice, i_start:i_stop]
                    sigs = sigs[channel_indexes]
                else:
                    sigs = sigs[channel_indexes, i_start:i_stop]
            else:
                if self._old_format:
                    sigs = sigs[self._channel_slice, i_start:i_stop]
                    sigs = sigs[sorted_channel_indexes]
                else:
                    sigs = sigs[sorted_channel_indexes, i_start:i_stop]
                sigs = sigs[resorted_indexes]
        except OSError as e:
            print("*" * 10)
            print(_hdf_maxwell_error)
            print("*" * 10)
            raise (e)
        sigs = sigs.T

        return sigs


_hdf_maxwell_error = """Maxwell file format is based on HDF5.
The internal compression requires a custom plugin!!!
This is a big pain for the end user.
You, as a end user, should ask Maxwell company to change this.
Please visit this page and install the missing decompression libraries:
https://share.mxwbio.com/d/4742248b2e674a85be97/
Then, link the decompression library by setting the `HDF5_PLUGIN_PATH` to your
installation location, e.g. via
os.environ['HDF5_PLUGIN_PATH'] = '/path/to/custom/hdf5/plugin/'

Alternatively, you can use the auto_install_maxwell_hdf5_compression_plugin() below
function that do it automagically.
"""


def auto_install_maxwell_hdf5_compression_plugin(hdf5_plugin_path=None, force_download=True):
    if hdf5_plugin_path is None:
        hdf5_plugin_path = os.getenv("HDF5_PLUGIN_PATH", None)
        if hdf5_plugin_path is None:
            hdf5_plugin_path = Path.home() / "hdf5_plugin_path_maxwell"
            os.environ["HDF5_PLUGIN_PATH"] = str(hdf5_plugin_path)
    hdf5_plugin_path = Path(hdf5_plugin_path)
    hdf5_plugin_path.mkdir(exist_ok=True)

    if platform.system() == "Linux":
        remote_lib = "https://share.mxwbio.com/d/4742248b2e674a85be97/files/?p=%2FLinux%2Flibcompression.so&dl=1"
        local_lib = hdf5_plugin_path / "libcompression.so"
    elif platform.system() == "Darwin":
        remote_lib = "https://share.mxwbio.com/d/4742248b2e674a85be97/files/?p=%2FMacOS%2Flibcompression.dylib&dl=1"
        local_lib = hdf5_plugin_path / "libcompression.dylib"
    elif platform.system() == "Windows":
        remote_lib = "https://share.mxwbio.com/d/4742248b2e674a85be97/files/?p=%2FWindows%2Fcompression.dll&dl=1"
        local_lib = hdf5_plugin_path / "compression.dll"

    if not force_download and local_lib.is_file():
        print(f"The h5 compression library for Maxwell is already located in {local_lib}!")
        return

    dist = urlopen(remote_lib)
    with open(local_lib, "wb") as f:
        f.write(dist.read())
